#!/usr/bin/python3

# Scale the final GDS by a factor of 4
# This is called by a tech hook that should be inserted post write_design

import sys, glob, os, re

def main():
    assert len(sys.argv) == 3, 'Must have 2 arguments: 1) Path to list of standard cells, 2) Path to GDS file'
    stdcells = sys.argv[1]
    gds_file = sys.argv[2]

    try:
        # Prioritize gdstk
        gds_tool = __import__('gdstk')
    except ImportError:
        try:
            print("gdstk not found, falling back to gdspy...")
            gds_tool = __import__('gdspy')
        except ImportError:
            print('Check your gdspy installation!')
            sys.exit()

    print('Scaling down {gds} using {tool}...'.format(gds=gds_file, tool=gds_tool.__name__))

    # load the standard cell list
    cell_list = [line.rstrip() for line in open(stdcells, 'r')]

    if gds_tool.__name__ == 'gdstk':
        # load original_gds
        gds_lib = gds_tool.read_gds(infile=gds_file)
        # Iterate through cells that aren't part of standard cell library and scale
        for cell in list(filter(lambda c: c.name not in cell_list, gds_lib.cells)):
            print('Scaling down ' + cell.name)

            # Need to remove 'blk' layer from any macros, else LVS rule deck interprets it as a polygon
            # This has a layer datatype of 4
            # Then scale down the polygon
            cell.filter(layers=[], types=[4], operation='or')
            for poly in cell.polygons:
                poly.scale(0.25)

            # Scale paths
            for path in cell.paths:
                path.scale(0.25)

            # Scale and move labels
            for label in cell.labels:
                # Bug fix for some EDA tools that didn't set MAG field in gds file
                # Maybe this is expected behavior in ASAP7 PDK
                label.magnification = 0.25
                label.origin = tuple(i * 0.25 for i in label.origin)

            # Move references (which are scaled if needed)
            for ref in cell.references:
                ref.origin = tuple(i * 0.25 for i in ref.origin)

        # We can also write an SVG of the top cell with gdstk
        gds_lib.top_level()[0].write_svg(os.path.splitext(gds_file)[0] + '.svg')

    elif gds_tool.__name__ == 'gdspy':
        # load original_gds
        gds_lib = gds_tool.GdsLibrary().read_gds(infile=gds_file, units='import')
        # Iterate through cells that aren't part of standard cell library and scale
        for k,v in gds_lib.cell_dict.items():
            if not any(cell in k for cell in cell_list):
                print('Scaling down ' + k)

                # Need to remove 'blk' layer from any macros, else LVS rule deck interprets it as a polygon
                # This has a layer datatype of 4
                # Then scale down the polygon
                v.polygons = [poly.scale(0.25) for poly in v.polygons if not 4 in poly.datatypes]

                # Scale paths
                for path in v.paths:
                    path.scale(0.25)
                    # gdspy v1.4 bug: we also need to scale custom path extensions
                    # Fixed by gdspy/pull#101 in next release
                    if gds_tool.__version__ == '1.4':
                        for i, end in enumerate(path.ends):
                            if isinstance(end, tuple):
                                path.ends[i] = tuple([e*0.25 for e in end])

                # Scale and move labels
                for label in v.labels:
                    # Bug fix for some EDA tools that didn't set MAG field in gds file
                    # Maybe this is expected behavior in ASAP7 PDK
                    # In gdspy/__init__.py: `kwargs['magnification'] = record[1][0]`
                    label.magnification = 0.25
                    label.translate(-label.position[0]*0.75, -label.position[1]*0.75)

                # Move references (which are scaled if needed)
                for ref in v.references:
                    ref.translate(-ref.origin[0]*0.75, -ref.origin[1]*0.75)

    # Overwrite original GDS file
    gds_lib.write_gds(gds_file)

    # For Calibre DRC, we need to rename the top cell to 'TOPCELL'
    gds_lib.top_level()[0].name = 'TOPCELL'
    gds_lib.write_gds(os.path.splitext(gds_file)[0] + '_drc.gds')



#  hammer-vlsi plugin for Cadence Genus.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerTool, HammerToolStep, HammerToolHookAction, HierarchicalMode
from hammer.utils import VerilogUtils
from hammer.vlsi import HammerSynthesisTool
from hammer.logging import HammerVLSILogging
from hammer.vlsi import MMMCCornerType
import hammer.tech as hammer_tech

from typing import Dict, List, Any, Optional

from hammer.tech.specialcells import CellType

import os
from collections import Counter

from hammer.common.cadence import CadenceTool


class Genus(HammerSynthesisTool, CadenceTool):
    @property
    def post_synth_sdc(self) -> Optional[str]:
        # No post-synth SDC input for synthesis...
        return None

    def fill_outputs(self) -> bool:
        # Check that the regs paths were written properly if the write_regs step was run
        self.output_seq_cells = self.all_cells_path
        self.output_all_regs = self.all_regs_path
        if self.ran_write_regs:
            if not os.path.isfile(self.all_cells_path):
                raise ValueError("Output find_regs_cells.json %s not found" % (self.all_cells_path))

            if not os.path.isfile(self.all_regs_path):
                raise ValueError("Output find_regs_paths.json %s not found" % (self.all_regs_path))

            if not self.process_reg_paths(self.all_regs_path):
                self.logger.error("Failed to process all register paths")
        else:
            self.logger.info("Did not run write_regs")

        # Check that the synthesis outputs exist if the synthesis run was successful
        mapped_v = self.mapped_hier_v_path if self.hierarchical_mode.is_nonleaf_hierarchical() else self.mapped_v_path
        self.output_files = [mapped_v]
        self.output_sdc = self.mapped_sdc_path
        self.sdf_file = self.output_sdf_path
        if self.ran_write_outputs:
            if not os.path.isfile(mapped_v):
                raise ValueError("Output mapped verilog %s not found" % (mapped_v)) # better error?

            if not os.path.isfile(self.mapped_sdc_path):
                raise ValueError("Output SDC %s not found" % (self.mapped_sdc_path)) # better error?

            if not os.path.isfile(self.output_sdf_path):
                raise ValueError("Output SDF %s not found" % (self.output_sdf_path))
        else:
            self.logger.info("Did not run write_outputs")

        return True

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = dict(super().env_vars)
        new_dict["GENUS_BIN"] = self.get_setting("synthesis.genus.genus_bin")
        return new_dict

    def export_config_outputs(self) -> Dict[str, Any]:
        outputs = dict(super().export_config_outputs())
        # TODO(edwardw): find a "safer" way of passing around these settings keys.
        outputs["synthesis.outputs.sdc"] = self.output_sdc
        outputs["synthesis.outputs.seq_cells"] = self.output_seq_cells
        outputs["synthesis.outputs.all_regs"] = self.output_all_regs
        outputs["synthesis.outputs.sdf_file"] = self.output_sdf_path
        return outputs

    def tool_config_prefix(self) -> str:
        return "synthesis.genus"

    def get_tool_hooks(self) -> List[HammerToolHookAction]:
        return [self.make_persistent_hook(genus_global_settings)]

    @property
    def steps(self) -> List[HammerToolStep]:
        steps_methods = [
            self.init_environment,
            self.syn_generic,
            self.syn_map,
            self.add_tieoffs,
            self.write_regs,
            self.generate_reports,
            self.write_outputs
        ]
        if self.get_setting("synthesis.inputs.retime_modules"):
            steps_methods.insert(1, self.retime_modules)
        return self.make_steps_from_methods(steps_methods)

    def do_pre_steps(self, first_step: HammerToolStep) -> bool:
        assert super().do_pre_steps(first_step)
        # Reload from the last checkpoint if we're not starting over.
        if first_step != self.first_step:
            self.verbose_append("read_db pre_{step}".format(step=first_step.name))
        return True

    def do_between_steps(self, prev: HammerToolStep, next: HammerToolStep) -> bool:
        assert super().do_between_steps(prev, next)
        # Write a checkpoint to disk.
        self.verbose_append("write_db -to_file pre_{step}".format(step=next.name))
        return True

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        return self.run_genus()

    @property
    def mapped_v_path(self) -> str:
        return os.path.join(self.run_dir, "{}.mapped.v".format(self.top_module))

    @property
    def mapped_hier_v_path(self) -> str:
        if self.version() >= self.version_number("191"):
            return os.path.join(self.run_dir, "{}_noilm.mapped.v".format(self.top_module))
        else:
            return os.path.join(self.run_dir, "genus_invs_des/genus.v.gz")

    @property
    def mapped_sdc_path(self) -> str:
        return os.path.join(self.run_dir, "{}.mapped.sdc".format(self.top_module))

    @property
    def all_regs_path(self) -> str:
        return os.path.join(self.run_dir, "find_regs_paths.json")

    @property
    def all_cells_path(self) -> str:
        return os.path.join(self.run_dir, "find_regs_cells.json")

    @property
    def output_sdf_path(self) -> str:
        return os.path.join(self.run_dir, "{top}.mapped.sdf".format(top=self.top_module))

    @property
    def ran_write_regs(self) -> bool:
        """The write_regs step sets this to True if it was run."""
        return self.attr_getter("_ran_write_regs", False)

    @ran_write_regs.setter
    def ran_write_regs(self, val: bool) -> None:
        self.attr_setter("_ran_write_regs", val)

    @property
    def ran_write_outputs(self) -> bool:
        """The write_outputs step sets this to True if it was run."""
        return self.attr_getter("_ran_write_outputs", False)

    @ran_write_outputs.setter
    def ran_write_outputs(self, val: bool) -> None:
        self.attr_setter("_ran_write_outputs", val)

    def remove_hierarchical_submodules_from_file(self, path: str) -> str:
        """
        Remove any hierarchical submodules' implementation from the given Verilog source file in path, if it is present.
        If it is not, return the original path.
        :param path: Path to verilog source file
        :return: A path to a modified version of the original file without the given module, or the same path as before.
        """
        with open(path, "r") as f:
            source = f.read()
        submodules = list(map(lambda ilm: ilm.module, self.get_input_ilms()))

        touched = False

        for submodule in submodules:
            if VerilogUtils.contains_module(source, submodule):
                source = VerilogUtils.remove_module(source, submodule)
                touched = True

        if touched:
            # Write the modified input to a new file in run_dir.
            name, ext = os.path.splitext(os.path.basename(path))
            new_filename = str(name) + "_no_submodules" + str(ext)
            new_path = os.path.join(self.run_dir, new_filename)
            with open(new_path, "w") as f:
                f.write(source)
            return new_path
        else:
            return path

    def init_environment(self) -> bool:
        # Python sucks here for verbosity
        verbose_append = self.verbose_append

        # Clock gating setup
        if self.get_setting("synthesis.clock_gating_mode") == "auto":
            verbose_append("set_db lp_clock_gating_infer_enable  true")
            # Innovus will create instances named CLKGATE_foo, CLKGATE_bar, etc.
            verbose_append("set_db lp_clock_gating_prefix  {CLKGATE}")
            verbose_append("set_db lp_insert_clock_gating  true")
            verbose_append("set_db lp_clock_gating_hierarchical true")
            verbose_append("set_db lp_insert_clock_gating_incremental true")
            verbose_append("set_db lp_clock_gating_register_aware true")

        # Set up libraries.
        # Read timing libraries.
        mmmc_path = os.path.join(self.run_dir, "mmmc.tcl")
        self.write_contents_to_path(self.generate_mmmc_script(), mmmc_path)
        verbose_append("read_mmmc {mmmc_path}".format(mmmc_path=mmmc_path))

        if self.hierarchical_mode.is_nonleaf_hierarchical():
            # Read ILMs.
            for ilm in self.get_input_ilms():
                # Assumes that the ILM was created by Innovus (or at least the file/folder structure).
                verbose_append("read_ilm -basename {data_dir}/{module}_postRoute -module_name {module}".format(
                    data_dir=ilm.data_dir, module=ilm.module))

        # Read LEF layouts.
        lef_files = self.technology.read_libs([
            hammer_tech.filters.lef_filter
        ], hammer_tech.HammerTechnologyUtils.to_plain_item)
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            ilm_lefs = list(map(lambda ilm: ilm.lef, self.get_input_ilms()))
            lef_files.extend(ilm_lefs)
        verbose_append("read_physical -lef {{ {files} }}".format(
            files=" ".join(lef_files)
        ))

        # Load input files and check that they are all Verilog.
        if not self.check_input_files([".v", ".sv"]):
            return False
        # We are switching working directories and Genus still needs to find paths.
        abspath_input_files = list(map(lambda name: os.path.join(os.getcwd(), name), self.input_files))  # type: List[str]

        # If we are in hierarchical, we need to remove hierarchical sub-modules/sub-blocks.
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            abspath_input_files = list(map(self.remove_hierarchical_submodules_from_file, abspath_input_files))

        # Add any verilog_synth wrappers (which are needed in some technologies e.g. for SRAMs) which need to be
        # synthesized.
        abspath_input_files += self.technology.read_libs([
            hammer_tech.filters.verilog_synth_filter
        ], hammer_tech.HammerTechnologyUtils.to_plain_item)

        # Read the RTL.
        verbose_append("read_hdl -sv {{ {} }}".format(" ".join(abspath_input_files)))

        # Elaborate/parse the RTL.
        verbose_append("elaborate {}".format(self.top_module))
        # Preserve submodules
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            for ilm in self.get_input_ilms():
                verbose_append("set_db module:{top}/{mod} .preserve true".format(top=self.top_module, mod=ilm.module))
        verbose_append("init_design -top {}".format(self.top_module))

        # Prevent floorplanning targets from getting flattened.
        # TODO: is there a way to track instance paths through the synthesis process?
        verbose_append("set_db root: .auto_ungroup none")

        # Set units to pF and technology time unit.
        # Must be done after elaboration.
        verbose_append("set_units -capacitance 1.0pF")
        verbose_append("set_load_unit -picofarads 1")
        verbose_append("set_units -time 1.0{}".format(self.get_time_unit().value_prefix + self.get_time_unit().unit))

        # Set "don't use" cells.
        for l in self.generate_dont_use_commands():
            self.append(l)

        return True

    def retime_modules(self) -> bool:
        retime_mods = self.get_setting("synthesis.inputs.retime_modules")

        if retime_mods:
            rt_tcl = (
                f"set rt_mods [get_designs \"{' '.join(retime_mods)}\"]\n" \
                "foreach rt_mod $rt_mods {\n" \
                "  set_db $rt_mod .retime true\n" \
                "}\n" \
                "set_db / .retime_verification_flow true"
            )
            self.append(rt_tcl)

        return True

    def syn_generic(self) -> bool:
        self.verbose_append("syn_generic")
        return True

    def syn_map(self) -> bool:
        self.verbose_append("syn_map")
        # Need to suffix modules for hierarchical simulation if not top
        if self.hierarchical_mode not in [HierarchicalMode.Flat, HierarchicalMode.Top]:
            self.verbose_append("update_names -module -log hier_updated_names.log -suffix _{MODULE}".format(MODULE=self.top_module))
        return True

    def add_tieoffs(self) -> bool:
        tie_hi_cells = self.technology.get_special_cell_by_type(CellType.TieHiCell)
        tie_lo_cells = self.technology.get_special_cell_by_type(CellType.TieLoCell)
        tie_hilo_cells = self.technology.get_special_cell_by_type(CellType.TieHiLoCell)

        if len(tie_hi_cells) != 1 or len (tie_lo_cells) != 1:
            if len(tie_hilo_cells) != 1:
                self.logger.warning("Hi and Lo tiecells are unspecified or improperly specified and will not be added during synthesis.")
                return True
            tie_hi_cells = tie_hilo_cells
            tie_lo_cells = tie_hilo_cells

        tie_hi_cell = tie_hi_cells[0].name[0]
        tie_lo_cell = tie_lo_cells[0].name[0]

        # Limit "no delay description exists" warnings
        self.verbose_append("set_db message:WSDF-201 .max_print 20")
        self.verbose_append("set_db use_tiehilo_for_const duplicate")

        # If there is more than 1 corner or a certain type, use lib cells for only the active analysis view
        corner_counts = Counter(list(map(lambda c: c.type, self.get_mmmc_corners())))
        if any(cnt>1 for cnt in corner_counts.values()):
            self.verbose_append("set ACTIVE_VIEW [string map { .setup_view {} .hold_view {} .extra_view {} } [get_db analysis_view:[get_analysis_views] .name]]")
            self.verbose_append("set HI_TIEOFF [get_db base_cell:{TIE_HI_CELL} .lib_cells -if {{ .library.default_opcond == $ACTIVE_VIEW }}]".format(TIE_HI_CELL=tie_hi_cell))
            self.verbose_append("set LO_TIEOFF [get_db base_cell:{TIE_LO_CELL} .lib_cells -if {{ .library.default_opcond == $ACTIVE_VIEW }}]".format(TIE_LO_CELL=tie_lo_cell))
            self.verbose_append("add_tieoffs -high $HI_TIEOFF -low $LO_TIEOFF -max_fanout 1 -verbose")
        else:
            self.verbose_append("add_tieoffs -high {HI_TIEOFF} -low {LO_TIEOFF} -max_fanout 1 -verbose".format(HI_TIEOFF=tie_hi_cell, LO_TIEOFF=tie_lo_cell))
        return True

    def generate_reports(self) -> bool:
        """Generate reports."""
        # TODO: extend report generation capabilities
        self.verbose_append("write_reports -directory reports -tag final")
        return True

    def write_regs(self) -> bool:
        """write regs info to be read in for simulation register forcing"""
        if self.hierarchical_mode.is_nonleaf_hierarchical():
            self.append(self.child_modules_tcl())
        self.append(self.write_regs_tcl())
        self.ran_write_regs = True
        return True

    def write_outputs(self) -> bool:
        verbose_append = self.verbose_append
        top = self.top_module

        verbose_append("write_hdl > {}".format(self.mapped_v_path))
        if self.hierarchical_mode.is_nonleaf_hierarchical() and self.version() >= self.version_number("191"):
            verbose_append("write_hdl -exclude_ilm > {}".format(self.mapped_hier_v_path))
        verbose_append("write_script > {}.mapped.scr".format(top))
        corners = self.get_mmmc_corners()
        if corners:
            # First setup corner is default view
            view_name="{cname}.setup_view".format(cname=next(filter(lambda c: c.type is MMMCCornerType.Setup, corners)).name)
        else:
            # TODO: remove hardcoded my_view string
            view_name = "my_view"
        verbose_append("write_sdc -view {view} > {file}".format(view=view_name, file=self.mapped_sdc_path))

        verbose_append("write_sdf > {run_dir}/{top}.mapped.sdf".format(run_dir=self.run_dir, top=top))

        # We just get "Cannot trace ILM directory. Data corrupted."
        # -hierarchical needs to be used for non-leaf modules
        is_hier = self.hierarchical_mode != HierarchicalMode.Leaf # self.hierarchical_mode != HierarchicalMode.Flat
        verbose_append("write_design -innovus {hier_flag} -gzip_files {top}".format(
            hier_flag="-hierarchical" if is_hier else "", top=top))

        self.ran_write_outputs = True

        return True

    def run_genus(self) -> bool:
        verbose_append = self.verbose_append

        """Close out the synthesis script and run Genus."""
        # Quit Genus.
        verbose_append("quit")

        # Create synthesis script.
        syn_tcl_filename = os.path.join(self.run_dir, "syn.tcl")
        self.write_contents_to_path("\n".join(self.output), syn_tcl_filename)

        # Build args.
        args = [
            self.get_setting("synthesis.genus.genus_bin"),
            "-f", syn_tcl_filename,
            "-no_gui"
        ]

        if bool(self.get_setting("synthesis.genus.generate_only")):
            self.logger.info("Generate-only mode: command-line is " + " ".join(args))
        else:
            # Temporarily disable colours/tag to make run output more readable.
            # TODO: think of a more elegant way to do this?
            HammerVLSILogging.enable_colour = False
            HammerVLSILogging.enable_tag = False
            self.run_executable(args, cwd=self.run_dir) # TODO: check for errors and deal with them
            HammerVLSILogging.enable_colour = True
            HammerVLSILogging.enable_tag = True

        return True

def genus_global_settings(ht: HammerTool) -> bool:
    """Settings that need to be reapplied at every tool invocation"""
    assert isinstance(ht, HammerSynthesisTool)
    assert isinstance(ht, CadenceTool)
    ht.create_enter_script()

    # Python sucks here for verbosity
    verbose_append = ht.verbose_append

    # Generic Settings
    verbose_append("set_db hdl_error_on_blackbox true")
    verbose_append("set_db max_cpus_per_server {}".format(ht.get_setting("vlsi.core.max_threads")))

    return True

tool = Genus



#  Older versions of readlink, like readlink 8.4, don't support multiple
#  arguments to readlink.
#  This script is meant to emulate that behaviour on those systems.
#  Usage:
#  readlink_array foo bar baz == readlink -f foo bar baz
#
#  See LICENSE for licence details.

import sys
import subprocess


def run(args):
    # Pop the script name.
    arg0 = args.pop(0)

    if len(args) == 0:
        print(arg0 + ": missing operand", file=sys.stderr)
        return 1

    for arg in sys.argv:
        # Run 'readlink -f' on each argument, and print the result.
        print(subprocess.check_output(["readlink", "-f", arg]).strip())

    return 0


def main():
    sys.exit(run(sys.argv))



#  Common code shared between all Vivado plugins.
#
#  See LICENSE for licence details.

from typing import Dict, List, Optional

import os
import shutil
from abc import ABCMeta

from hammer.utils import deepdict
from hammer.vlsi import HammerTool


class VivadoCommon(HammerTool, metaclass=ABCMeta):
    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = deepdict(super().env_vars)
        return new_dict

    def append(self, cmd: str) -> None:
        self.tcl_append(cmd, self.output)

    def setup_workspace(self) -> bool:
        # make object directory
        os.makedirs(os.path.join(self.run_dir, 'obj'), exist_ok=True)
        # copy constraint file
        cons_fname = os.path.abspath(
            self.get_setting('synthesis.vivado.constraints_file'))
        cons_dir = os.path.join(self.run_dir, 'constrs')
        os.makedirs(cons_dir, exist_ok=True)
        cons_targ = os.path.join(cons_dir, os.path.basename(cons_fname))
        shutil.copyfile(cons_fname, cons_targ)

        self.output = []  # type: List[str]
        return True

    def get_file_contents(self, file_name: str,
                          file_params: Optional[Dict[str, str]]) -> str:
        if os.path.isabs(file_name):
            fname = file_name  # type: str
        else:
            raise RuntimeError("No clue what's going on here - tool_dir no longer exists, this plugin should use Python resources")
            # fname = os.path.join(self.tool_dir, 'file_templates', file_name)
        with open(fname, 'r') as f:
            content = f.read()
            if file_params:
                content = content.format(**file_params)
            return content

    def append_file(self, file_name: str, file_params: Optional[Dict[str, str]]) -> None:
        for line in self.get_file_contents(file_name,
                                           file_params).splitlines():
            self.append(line)

    def generate_board_defs(self) -> bool:
        file_params = {
            'board_name': self.get_setting('synthesis.vivado.board_name'),
            'part_fpga': self.get_setting('synthesis.vivado.part_fpga'),
            'part_board': self.get_setting('synthesis.vivado.part_board'),
        }
        self.append_file('board.tcl', file_params)
        return True

    def generate_paths_and_src_defs(self) -> bool:
        verilog_files = ' '.join((os.path.abspath(fname)
                                  for fname in self.input_files
                                  if fname.endswith('.v')))
        file_params = {
            'board_files':
            self.get_setting('synthesis.vivado.board_files') or '""',
            'dcp_macro_dir':
            self.get_setting('synthesis.vivado.dcp_macro_dir') or '""',
            'work_dir':
            self.run_dir,
            'verilog_files':
            verilog_files,
            'top_module':
            self.top_module,
        }
        self.append_file('paths.tcl', file_params)
        return True

    def generate_project_defs(self) -> bool:
        file_params = {
            'part_fpga': self.get_setting('synthesis.vivado.part_fpga'),
            'part_board': self.get_setting('synthesis.vivado.part_board'),
        }
        self.append_file('project.tcl', file_params)
        return True

    def generate_run_script(self, script_name: str,
                            file_params: Dict[str, str]) -> str:
        content = self.get_file_contents(script_name, file_params)
        fpath = os.path.join(self.run_dir, script_name)
        with open(fpath, 'w') as f:
            f.write(content)
        os.chmod(fpath, 0o755)
        return fpath



#  hammer-generate-mdf
#  Helper script to generate an MDF description of a memory from a
#  ReplSeqMem config file.
#
#  See LICENSE for licence details.

import argparse
from collections import namedtuple
import json
import re
import sys

MEM_REGEX = r"name ([^\s]+) depth (\d+) width (\d+) ports ((?:[^\s]+?\s?(?!mask_gran))+)\s?(?:mask_gran (\d+))?"

Mem = namedtuple('Mem', 'name depth width ports mask_gran')

def parseLine(line):
    """
    Parse a line using the given regex above or return None.
    """
    match = re.match(MEM_REGEX, line)
    if match is None:
        return None

    name = match.groups()[0]
    depth = int(match.groups()[1])
    width = int(match.groups()[2])
    ports = match.groups()[3].split(',')
    mask_gran = match.groups()[4]
    if mask_gran is not None:
        mask_gran = int(mask_gran)
    result = Mem(name=name, depth=depth, width=width, ports=ports, mask_gran=mask_gran)
    return result

def memToJSON(mem):
    result = {}
    result['type'] = "sram"
    result['name'] = mem.name
    result['depth'] = mem.depth
    result['width'] = mem.width
    result['ports'] = []
    num_read = -1
    num_write = -1
    num_rw = -1
    for port_type in mem.ports:
        port = {}
        port_name = "" # e.g. "RW0"
        if port_type == "read":
            num_read += 1
            port_name = "R%d" % num_read
        elif port_type.endswith("write"):
            num_write += 1
            port_name = "W%d" % num_write
        elif port_type.endswith("rw"):
            num_rw += 1
            port_name = "RW%d" % num_rw
        else:
            raise ValueError("Invalid port type " + port_type)

        port['clock port name'] = port_name + "_clk"
        port['address port name'] = port_name + "_addr"
        port['chip enable port name'] = port_name + "_en"

        if port_type == "read":
            port['output port name'] = port_name + "_data"
        elif "RW" in port_name:
            port['output port name'] = port_name + "_rdata"

        if port_type.endswith("write"):
            port['input port name'] = port_name + "_data"
        elif "RW" in port_name:
            port['input port name'] = port_name + "_wdata"
            port['write enable port name'] = port_name + "_wmode"

        if mem.mask_gran is not None:
            port['mask granularity'] = mem.mask_gran
            if "RW" in port_name:
                port['mask port name'] = port_name + "_wmask"
            else:
                port['mask port name'] = port_name + "_mask"

        result['ports'].append(port)
    return result

def run(args):
    try:
        lines = [line.strip() for line in open(args.macros)]
    except IOError:
        print("Could not read macros file: " + args.macros)
        return 1

    json_objs = []

    for line in lines:
        mem = parseLine(line)
        if mem is None:
            print('Failed to parse line "' + line + '" in ' + args.macros)
            return 1
        json_objs.append(memToJSON(mem))

    json_str = json.dumps(json_objs, sort_keys=True, indent=4)
    try:
        f = open(args.output, 'w')
        f.write(json_str)
    except IOError:
        print("Could not write output file: " + args.output)
        return 1

    return 0

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--macros", required=True, help="Config file generated by ReplSeqMem")
    parser.add_argument("--output", required=True, help="MDF output for MacroCompiler")
    args = parser.parse_args()
    sys.exit(run(args))




import os, tempfile, subprocess
from pathlib import Path

from hammer.vlsi import MMMCCorner, MMMCCornerType, HammerTool, HammerToolStep, HammerSRAMGeneratorTool, SRAMParameters
from hammer.tech import Corner, Supplies, Provide
from hammer.vlsi.units import VoltageValue, TemperatureValue
from hammer.tech import Library, ExtraLibrary
from typing import NamedTuple, Dict, Any, List
from abc import ABCMeta, abstractmethod

class SKY130SRAMGenerator(HammerSRAMGeneratorTool):
    def tool_config_prefix(self) -> str:
        return "sram_generator.sky130"

    def version_number(self, version: str) -> int:
        return 0

    # Run generator for a single sram and corner
    def generate_sram(self, params: SRAMParameters, corner: MMMCCorner) -> ExtraLibrary:
        cache_dir = os.path.abspath(self.technology.cache_dir)

        #TODO: this is really an abuse of the corner stuff
        if corner.type == MMMCCornerType.Setup:
            speed_name = "slow"
            speed = "SS"
            # self.logger.error("SKY130 SRAM cache does not support corner: {}".format(speed_name))
        elif corner.type == MMMCCornerType.Hold:
            speed_name = "fast"
            speed = "FF"
            # self.logger.error("SKY130 SRAM cache does not support corner: {}".format(speed_name))
        elif corner.type == MMMCCornerType.Extra:
            speed_name = "typical"
            speed = "TT"

        if params.family != "1rw" and params.family != "1rw1r":
            self.logger.error("SKY130 SRAM cache does not support family:{f}".format(f=params.family))
            return ExtraLibrary(prefix=None, library=None)  # type: ignore

        if params.name.startswith("sramgen_sram"):
            self.logger.info(f"Compiling {params.family} memories to SRAM22 instances")
            # s=round(round(params.width*params.depth/8, -3)/1000) # size in kiB
            w=params.width
            d=params.depth
            sram_name = params.name
            #TODO: replace this if SRAM22 characterization done for other corners
            # we only have typical lib for sky130 srams
            corner_str = "tt_025C_1v80"
            #corner_str = "{speed}_{volt}V_{temp}C".format(
            #        speed = speed,
            #        volt = str(corner.voltage.value_in_units("V")).replace(".","p"),
            #        temp = str(int(corner.temp.value_in_units("C"))).replace(".","p"))

            base_dir=self.get_setting('technology.sky130.sram22_sky130_macros')
            lib_path="{b}/{n}/{n}_{c}.lib".format(b=base_dir,n=sram_name,c=corner_str)
            if not os.path.exists(lib_path):
                self.logger.error(f"SKY130 {params.family} SRAM cache does not support corner: {corner_str}")
            
            return ExtraLibrary(prefix=None, library=Library(
                name=sram_name,
                nldm_liberty_file=lib_path,
                lef_file="{b}/{n}/{n}.lef".format(b=base_dir,n=sram_name),
                gds_file="{b}/{n}/{n}.gds".format(b=base_dir,n=sram_name),
                spice_file="{b}/{n}/{n}.spice".format(b=base_dir,n=sram_name),
                verilog_sim="{b}/{n}/{n}.v".format(b=base_dir,n=sram_name),
                corner=Corner(nmos=speed_name, pmos=speed_name, temperature=str(corner.temp.value_in_units("C")) + " C"),
                supplies=Supplies(VDD=str(corner.voltage.value_in_units("V")) + " V", GND="0 V"),
                provides=[Provide(lib_type="sram", vt=params.vt)]))

        # TODO: remove OpenRAM support very soon
        elif params.name.startswith("sky130_sram_"):
            self.logger.info(f"Compiling {params.family} memories to OpenRAM instances")
            base_dir = self.get_setting("technology.sky130.openram_lib")
            s=round(round(params.width*params.depth/8, -3)/1000) # size in kiB
            w=params.width
            d=params.depth
            m=8
            sram_name = f"sky130_sram_{s}kbyte_{params.family}_{w}x{d}_{m}"
            #TODO: Hammer SRAMParameters doesn't have this info
            #TODO: replace this if OpenRAM characterization done for other corners
            # we only have typical lib for sky130 srams
            corner_str = "TT_1p8V_25C"
            #corner_str = "{speed}_{volt}V_{temp}C".format(
            #        speed = speed,
            #        volt = str(corner.voltage.value_in_units("V")).replace(".","p"),
            #        temp = str(int(corner.temp.value_in_units("C"))).replace(".","p"))
            lib_path = "{b}/{n}/{n}_{c}.lib".format(
                b=base_dir,
                n=sram_name,
                c=corner_str)
            if not os.path.exists(lib_path):
                self.logger.error(f"SKY130 {params.family} SRAM cache does not support corner: {corner_str}")

            self.setup_openram_spice(sram_name)
            self.setup_openram_lef(sram_name)
            self.setup_openram_verilog(sram_name)
            # self.setup_sram_lib(sram_name)
            return ExtraLibrary(prefix=None, library=Library(
                name=sram_name,
                nldm_liberty_file="{b}/{n}/{n}_{c}.lib".format(b=base_dir,n=sram_name,c=corner_str),
                lef_file="{b}/{n}/{n}.lef".format(b=cache_dir,n=sram_name),
                gds_file="{b}/{n}/{n}.gds".format(b=base_dir,n=sram_name),
                spice_file="{b}/{n}/{n}.lvs.sp".format(b=cache_dir,n=sram_name),
                verilog_sim="{b}/{n}/{n}.v".format(b=cache_dir,n=sram_name),
                corner=Corner(nmos=speed_name, pmos=speed_name, temperature=str(corner.temp.value_in_units("C")) + " C"),
                supplies=Supplies(VDD=str(corner.voltage.value_in_units("V")) + " V", GND="0 V"),
                provides=[Provide(lib_type="sram", vt=params.vt)]))
        else:
            self.logger.error(f"SRAM {params.name} not supported")
            return ExtraLibrary(prefix=None, library=Library())


    def setup_openram_spice(self,sram_name) -> None:
        source_path = Path(self.get_setting("technology.sky130.openram_lib")) / sram_name / f"{sram_name}.lvs.sp"
        dest_path = f"{os.path.abspath(self.technology.cache_dir)}/{sram_name}/{sram_name}.lvs.sp"
        self.technology.ensure_dirs_exist(dest_path)
        if not source_path.exists():
            raise FileNotFoundError(f"SRAM Spice file not found: {source_path}")        
        with open(source_path,'r') as sf:
            with open(dest_path,'w') as df:
                self.logger.info("Modifying SRAM SPICE file: {} -> {}".format
                    (source_path, dest_path))
                for line in sf:
                    line = line.replace('sky130_fd_pr__pfet_01v8','pshort')
                    line = line.replace('sky130_fd_pr__nfet_01v8','nshort')
                    if sram_name == "sky130_sram_1kbyte_1rw1r_8x1024_8":
                        line = line.replace('wmask0[0]'    , 'wmask0')
                    df.write(line)


    def setup_openram_lef(self,sram_name) -> None:
        source_path = Path(self.get_setting("technology.sky130.openram_lib")) / sram_name / f"{sram_name}.lef"
        dest_path = f"{os.path.abspath(self.technology.cache_dir)}/{sram_name}/{sram_name}.lef"
        self.technology.ensure_dirs_exist(dest_path)
        if not source_path.exists():
            raise FileNotFoundError(f"SRAM LEF file not found: {source_path}")
        with open(source_path,'r') as sf:
            with open(dest_path,'w') as df:
                self.logger.info("Modifying SRAM LEF file: {} -> {}".format
                    (source_path, dest_path))
                units=False
                for line in sf:
                    if line.strip().startswith("UNITS"):
                        units=True
                    if line.strip().startswith("END UNITS"):
                        units=False
                        continue
                    if not units:
                        df.write(line)


    def setup_openram_verilog(self, sram_name) -> None:
        """ Move 'mem' declaration before it is referenced in the verilog. """
        source_path = Path(self.get_setting("technology.sky130.openram_lib")) / sram_name / f"{sram_name}.v"
        dest_path = f"{os.path.abspath(self.technology.cache_dir)}/{sram_name}/{sram_name}.v"
        if not source_path.exists():
            raise FileNotFoundError(f"SRAM Spice file not found: {source_path}")
        self.technology.ensure_dirs_exist(dest_path)
        with open(source_path,'r') as sf:
            with open(dest_path,'w') as df:
                self.logger.info("Modifying SRAM Verilog file: {} -> {}".format
                    (source_path, dest_path))
                lines = sf.readlines()
                insert_idx = 0
                for i,line in enumerate(lines):
                    if insert_idx == 0 and line.strip().startswith('always'):
                        insert_idx = i
                    elif line.strip() == "reg [DATA_WIDTH-1:0]    mem [0:RAM_DEPTH-1];":
                        lines.pop(i)
                        lines.insert(insert_idx,line)
                df.write(''.join(lines))

    def setup_sram_lib(self, sram_name) -> None:
        """ Flip endianness of SRAM ports. """
        source_path = Path(self.get_setting("technology.sky130.openram_lib")) / sram_name / f"{sram_name}_TT_1p8V_25C.lib"
        dest_path = f"{os.path.abspath(self.technology.cache_dir)}/{sram_name}/{sram_name}.v"
        if not source_path.exists():
            raise FileNotFoundError(f"SRAM Lib file not found: {source_path}")
        self.technology.ensure_dirs_exist(dest_path)
        with open(source_path,'r') as sf:
            with open(dest_path,'w') as df:
                self.logger.info("Modifying SRAM Lib file: {} -> {}".format
                    (source_path, dest_path))
                lines = sf.readlines()
                insert_idx = 0
                bit_from_line, bit_to_line = None,None
                # swap_bits = False
                for i,line in enumerate(lines):
                    if line.strip().startswith("bit_from"):
                        bit_from_line = line
                    elif line.strip().startswith("bit_to"):
                        bit_to_line = line
                    else:
                        df.write(line)
                    if bit_from_line is not None and bit_to_line is not None:
                        bit_from = bit_from_line.strip().replace(':','').replace(';','').split()[1]
                        bit_to   = bit_to_line.strip().replace(':','').replace(';','').split()[1]
                        if int(bit_from) > int(bit_to):
                            bit_from_line = bit_from_line.replace(bit_from,bit_to)
                            bit_to_line = bit_to_line.replace(bit_to,bit_from)
                        df.write(bit_from_line)
                        df.write(bit_to_line)
                        bit_from_line, bit_to_line = None,None

tool=SKY130SRAMGenerator



#  units.py
#  Unit classes/functions for hammer_vlsi.
#
#  See LICENSE for licence details.

from abc import abstractmethod
import sys
try:
    from abc import ABC  # pylint: disable=ungrouped-imports
except ImportError:
    if sys.version_info.major == 3 and sys.version_info.minor < 4:
        # Python compatibility: 3.3
        # Python 3.3 and below don't have abc.ABC
        import abc  # pylint: disable=ungrouped-imports
        ABC = abc.ABCMeta('ABC', (object,), {'__slots__': ()})  # type: ignore

from typing import Optional, TypeVar

from hammer.utils import get_or_else

_TT = TypeVar('_TT', bound='ValueWithUnit')

class ValueWithUnit(ABC):
    """Represents some particular value that has units (e.g. "10 ns", "2000 um", "25 C", etc).
    """

    # From https://stackoverflow.com/a/10970888
    _prefix_table = {
        'y': 1e-24,  # yocto
        'z': 1e-21,  # zepto
        'a': 1e-18,  # atto
        'f': 1e-15,  # femto
        'p': 1e-12,  # pico
        'n': 1e-9,  # nano
        'u': 1e-6,  # micro
        'm': 1e-3,  # milli
        'c': 1e-2,  # centi
        'd': 1e-1,  # deci
        '':  1,    # <no prefix>
        'k': 1e3,  # kilo
        'M': 1e6,  # mega
        'G': 1e9,  # giga
        'T': 1e12,  # tera
        'P': 1e15,  # peta
        'E': 1e18,  # exa
        'Z': 1e21,  # zetta
        'Y': 1e24,  # yotta
    }

    @property
    @abstractmethod
    def unit(self) -> str:
        """Get the base unit for values (e.g. "s", "m", "V", etc).
        Meant to be overridden by subclasses."""

    @property
    @abstractmethod
    def unit_type(self) -> str:
        """Get the base unit type for values. (e.g. for "s", this would be "time")
        Meant to be overridden by subclasses."""

    @property
    @abstractmethod
    def default_prefix(self) -> str:
        """Get the default prefix for values.
        (e.g. for time, specifying "n" would mean "0.25" would be interpreted as "0.25 ns".)
        Meant to be overridden by subclasses."""

    def __init__(self, value: str, prefix: Optional[str] = None) -> None:
        """
        Create a value from parsing the given string.
        :param value: Value encoded in the given string.
        :param prefix: If value does not have a prefix (e.g. "0.25"), then use
                       the given prefix, or the default prefix defined by the
                       class if one is not specified.
        """
        import re

        default_prefix = get_or_else(prefix, self.default_prefix)

        regex = r"^(-?[\d.]+) *(.*){}$".format(re.escape(self.unit))
        match = re.search(regex, value)
        if match is None:
            try:
                num = str(float(value))
                self._value_prefix = default_prefix
            except ValueError:
                raise ValueError("Malformed {type} value {value}".format(type=self.unit_type,
                                                                         value=value))
        else:
            num = match.group(1)
            self._value_prefix = match.group(2)

        if num.count('.') > 1 or len(self._value_prefix) > 1:
            raise ValueError("Malformed {type} value {value}".format(type=self.unit_type,
                                                                     value=value))

        if self._value_prefix not in self._prefix_table:
            raise ValueError("Bad prefix for {value}".format(value=value))

        self._value = float(num)  # type: float
        # Preserve the prefix too to preserve precision
        self._prefix = self._prefix_table[self._value_prefix]  # type: float

    @property
    def value_prefix(self) -> str:
        """Get the prefix string of this value."""
        return self._value_prefix

    @property
    def value(self) -> float:
        """Get the actual value of this value. (e.g. 10 ns -> 1e-9)"""
        return self._value * self._prefix

    def value_in_units(self, prefix: str, round_zeroes: bool = True) -> float:
        """Get this value in the given prefix. e.g. "ns", "mV", etc.
        """
        # e.g. extract "n" from "ns" or blank if it's blank (e.g. "V" -> "")
        letter_prefix = ""
        if prefix != self.unit:
            letter_prefix = "" if prefix == "" else prefix[0]

        retval = self._value * (self._prefix / self._prefix_table[letter_prefix])
        if round_zeroes:  # pylint: disable=no-else-return
            return round(retval, 3)
        else:
            return retval

    def str_value_in_units(self, prefix: str, round_zeroes: bool = True) -> str:
        """Get this value in the given prefix but including the units.
        e.g. return "5 ns".

        :param prefix: Prefix for the resulting value - e.g. "ns".
        :param round_zeroes: True to round 1.00000001 etc to 1 within 3 decimal places.
        """
        # %g removes trailing zeroes
        return "%g" % (self.value_in_units(prefix, round_zeroes)) + " " + prefix

    # Comparison operators.
    # Note that mypy doesn't properly support type checking on equality
    # operators so the type of __eq__ is object :(
    # As a result, the operators' (e.g. __eq__) 'other' type can't be _TT.
    # Therefore, we implement the operators themselves separately and then wrap
    # them in the special operators.
    # See https://github.com/python/mypy/issues/1271
    # Disable useless pylint checks for the following methods.
    # pylint: disable=unidiomatic-typecheck

    def eq(self: _TT, other: _TT) -> bool:  # pylint: disable=invalid-name
        """
        Compare equality of this value with another.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return self.value_in_units(self.default_prefix) == other.value_in_units(self.default_prefix)

    def __eq__(self: _TT, other: object) -> bool:
        """
        Compare equality of this value with another.
        The types must match.
        """
        return self.eq(other)  # type: ignore

    def ne(self: _TT, other: _TT) -> bool:  # pylint: disable=invalid-name
        """
        Compare inequality of this value with another.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return not self.eq(other)

    def __ne__(self: _TT, other: object) -> bool:
        """
        Compare inequality of this value with another.
        The types must match.
        """
        return self.ne(other)  # type: ignore

    def __lt__(self: _TT, other: _TT) -> bool:
        """
        Check if self is less than other.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return self.value < other.value

    def __le__(self: _TT, other: _TT) -> bool:
        """
        Check if self is less than or equal to other.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return self.value <= other.value

    def __gt__(self: _TT, other: _TT) -> bool:
        """
        Check if self is greater than other.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return self.value > other.value

    def __ge__(self: _TT, other: _TT) -> bool:
        """
        Check if self is greater than or equal to other.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return self.value >= other.value

    def __add__(self: _TT, other: _TT) -> _TT:
        """
        Add other and self.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return type(self)(str(self.value + other.value),"")

    def __sub__(self: _TT, other: _TT) -> _TT:
        """
        Subtract other from self.
        The types must match.
        """
        if type(self) != type(other):
            raise TypeError("Types do not match")
        return type(self)(str(self.value - other.value),"")

    def __div__(self: _TT, other: float) -> _TT:
        """
        Divide self by a float or an integer.
        """
        raise NotImplementedError()

    # Some python nonsense
    def __truediv__(self: _TT, other: float) -> _TT:
        return type(self)(str(self.value / other),"")

    def __mul__(self: _TT, other: float) -> _TT:
        """
        Multiply self by a float or an integer.
        """
        return type(self)(str(self.value * other),"")


class TimeValue(ValueWithUnit):
    """Time value - e.g. "4 ns".
    Parses time values from strings.
    """

    @property
    def default_prefix(self) -> str:
        """Default prefix: ns"""
        return "n"

    @property
    def unit(self) -> str:
        return "s"

    @property
    def unit_type(self) -> str:
        return "time"


class VoltageValue(ValueWithUnit):
    """Voltage value - e.g. "0.95 V", "950 mV".
    """

    @property
    def default_prefix(self) -> str:
        """Default is plain volts (e.g. "0.1" -> 0.1 V)."""
        return ""

    @property
    def unit(self) -> str:
        return "V"

    @property
    def unit_type(self) -> str:
        return "voltage"


class TemperatureValue(ValueWithUnit):
    """Temperature value in Celsius - e.g. "25 C", "125 C".
    Mainly used for specifying corners for MMMC.
    """

    @property
    def default_prefix(self) -> str:
        """Default is plain degrees Celsius (e.g. "25" -> "25 C")."""
        return ""

    @property
    def unit(self) -> str:
        return "C"

    @property
    def unit_type(self) -> str:
        return "voltage"


class CapacitanceValue(ValueWithUnit):
    """Capacitance value - e.g. "5 fF", "10 nF".
    """

    @property
    def default_prefix(self) -> str:
        """Default prefix: fF"""
        return "f"

    @property
    def unit(self) -> str:
        return "F"

    @property
    def unit_type(self) -> str:
        return "capacitance"



#  hammer-vlsi plugin for Cadence Conformal.
#
#  See LICENSE for licence details.

from typing import List, Dict, Tuple

import os
import errno

from hammer.vlsi import HammerFormalTool, HammerToolStep
from hammer.logging import HammerVLSILogging
import hammer.tech as hammer_tech
from hammer.common.cadence import CadenceTool

# Notes: Tcl mode is enabled for harmonization with other Cadence tools and additional Tcl functionality.
# There is a minor performance hit with database operations vs. native language.

class Conformal(HammerFormalTool, CadenceTool):

    def tool_config_prefix(self) -> str:
        return "formal.conformal"

    @property
    def env_vars(self) -> Dict[str, str]:
        v = dict(super().env_vars)
        if self.check in ["constraint", "cdc"]:
            v["CONFORMAL_BIN"] = self.get_setting("formal.conformal.conformal_ccd_bin")
        else:
            v["CONFORMAL_BIN"] = self.get_setting("formal.conformal.conformal_lec_bin")
        return v

    @property
    def start_cmd(self) -> List[str]:
        """ Generate required startup command based on the requested check and license level """
        lec_bin = self.get_setting("formal.conformal.conformal_lec_bin")
        ccd_bin = self.get_setting("formal.conformal.conformal_ccd_bin")
        license = self.get_setting("formal.conformal.license")
        cmd = ["", ""]
        if not license in ["L", "XL", "GXL"]:
            self.logger.error("License must be L, XL, or GXL. For CCD, -MCC is equivalent to GXL here.")

        if self.check == "lec":
            cmd = [lec_bin, f"-{license}"]
        elif self.check == "power":
            if license == "L":
                self.logger.error("power not supported with L license")
            else:
                cmd = [lec_bin, f"-LP{license}"]
        elif self.check == "eco":
            if license == "L":
                self.logger.error("eco not supported with L license")
            elif license == "XL":
                cmd = [lec_bin, "-ECO"]
            else:
                cmd = [lec_bin, "-ECOGXL"]
        elif self.check == "property":
            return [lec_bin, "-VERIFY"]
        elif self.check in ["constraint", "cdc"]:
            if license == "GXL":
                cmd = [ccd_bin, "-MCC"]
            else:
                cmd = [ccd_bin, f"-{license}"]
        else:
            self.logger.error("Unsupported check type")

        return cmd

    def check_reference_files(self, extensions: List[str]) -> bool:
        """
        Verify that reference files exist and have the specified extensions.
        Analogous to check_input_files in HammerTool.

        :param extensions: List of extensions e.g. [".v", ".sv"]
        :return: True if all files exist and have the specified extensions.
        """
        refs = self.reference_files
        error = False
        for r in refs:
            if not r.endswith(tuple(extensions)):
                self.logger.error(f"Input of unsupported type {r} detected!")
                error = True
            if not os.path.isfile(r):
                self.logger.error(f"Input file {r} does not exist!")
                error = True
        return not error

    @property
    def restart_checkpoint(self) -> str:
        """ Name of checkpoint to be restarted from (set by do_pre_steps) """
        return self.attr_getter("_restart_checkpoint", "")

    @restart_checkpoint.setter
    def restart_checkpoint(self, val: str) -> None:
        self.attr_setter("_restart_checkpoint", val)

    @property
    def _step_transitions(self) -> List[Tuple[str, str]]:
        """
        Private helper property to keep track of which steps we ran so that we
        can create symlinks.
        This is a list of (pre, post) steps
        """
        return self.attr_getter("__step_transitions", [])

    @_step_transitions.setter
    def _step_transitions(self, value: List[Tuple[str, str]]) -> None:
        self.attr_setter("__step_transitions", value)


    def do_pre_steps(self, first_step: HammerToolStep) -> bool:
        assert super().do_pre_steps(first_step)
        # Restart from the last checkpoint if we're not starting over.
        # Not in the dofile, must be a command-line option
        if first_step != self.first_step:
            self.restart_checkpoint = f"pre_{first_step.name}"
        return True

    def do_between_steps(self, prev: HammerToolStep, next: HammerToolStep) -> bool:
        assert super().do_between_steps(prev, next)
        # Write a checkpoint to disk.
        self.append(f"checkpoint pre_{next.name} -replace")
        # Symlink the checkpoint to latest for open_checkpoint script later.
        self.append(f"ln -sfn pre_{next.name} latest")
        self._step_transitions = self._step_transitions + [(prev.name, next.name)]
        return True

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        # Create symlinks for post_<step> to pre_<step+1> to improve usability.
        try:
            for prev, next in self._step_transitions:
                os.symlink(
                    os.path.join(self.run_dir, f"pre_{next}"), # src
                    os.path.join(self.run_dir, f"post_{prev}") # dst
                )
        except OSError as e:
            if e.errno != errno.EEXIST:
                self.logger.warning("Failed to create post_* symlinks: " + str(e))

        # Create checkpoint post_<last step>
        # TODO: this doesn't work if you're only running the very last step
        if len(self._step_transitions) > 0:
            last = f"post_{self._step_transitions[-1][1]}"
            self.append(f"checkpoint {last} -replace")
            # Symlink the database to latest for open_checkpoint script later.
            self.append(f"ln -sfn {last} latest")

        return self.generate_open_checkpoint() and self.run_conformal()

    @property
    def steps(self) -> List[HammerToolStep]:
        if self.check != "lec":
            self.logger.error("Check type {self.check} not yet supported!")
        steps = [
            self.setup_designs,
            self.compare_designs
        ]
        return self.make_steps_from_methods(steps)

    def setup_designs(self) -> bool:
        """ Setup the designs """
        append = self.append

        # Exit on dofile error
        append("set_dofile_abort exit")

        # Multithreading (max 16 allowed by tool)
        max_threads = min(self.get_setting("vlsi.core.max_threads"), 16)
        append(f"set_parallel_option -threads 1,{max_threads}")

        # Read libraries (macros, stdcells)
        # TODO: support VHDL + Liberty. For now, -sva = SystemVerilog w/ assertion support.
        lib_v_files = self.technology.read_libs(
                [hammer_tech.filters.verilog_synth_filter],
                hammer_tech.HammerTechnologyUtils.to_plain_item)
        lib_v_files.extend(self.technology.read_libs(
                [hammer_tech.filters.verilog_sim_filter],
                hammer_tech.HammerTechnologyUtils.to_plain_item))
        append(f"read_library {' '.join(lib_v_files)} -sva -bboxsolver -both")

        # Read designs
        valid_exts = [".v", ".v.gz", ".sv", ".sv.gz"]
        if not self.check_input_files(valid_exts) or not self.check_reference_files(valid_exts):
            return False
        golden_files = list(map(lambda name: os.path.join(os.getcwd(), name), self.reference_files))
        append(f"read_design {' '.join(golden_files)} -sva -golden")
        revised_files = list(map(lambda name: os.path.join(os.getcwd(), name), self.input_files))
        append(f"read_design {' '.join(revised_files)} -sva -revised")

        # Set top module
        append(f"set_root_module {self.top_module} -both")

        # Auto setup analysis optimizations
        if self.get_setting("formal.conformal.license") != "L":
            append("set_analyze_option -auto")

        # Setup reports
        append("report_design_data")

        return True

    def compare_designs(self) -> bool:
        """ Depending on license, performs flat or hierarchical comparison """
        append = self.append

        if self.get_setting("formal.conformal.license") == "L":
            append("report_black_box")
            append("set_system_mode lec")
            append("add_compare_point -all")
            append("compare")
            append("report_compare_data")
        else:
            # TODO: need resource file for DC-mapped netlists
            append('write_hier_compare_dofile hier_compare.tcl -replace '\
                   '-prepend_string "analyze_datapath -module; analyze_datapath"')
            append("run_hier_compare hier_compare.tcl")
            append("set_system_mode lec")

        append("report_statistics")

        return True

    def generate_open_checkpoint(self) -> bool:
        # Make sure that generated-scripts exists.
        generated_scripts_dir = os.path.join(self.run_dir, "generated-scripts")
        os.makedirs(generated_scripts_dir, exist_ok=True)

        # Script to open results checkpoint
        self.create_enter_script()
        open_checkpoint_tcl = os.path.join(generated_scripts_dir, "open_checkpoint.tcl")
        with open(open_checkpoint_tcl, "w") as f:
            f.write("set_gui -mapping")
        open_checkpoint_script = os.path.join(generated_scripts_dir, "open_checkpoint")
        with open(open_checkpoint_script, "w") as f:
            assert super().do_pre_steps(self.first_step)
            args = self.start_cmd
            args.extend(["-gui", "-restart_checkpoint", "latest"])
            f.write("#!/bin/bash\n")
            f.write(f"cd {self.run_dir}\n")
            f.write("source enter\n")
            f.write(f"$CONFORMAL_BIN -restart_checkpoint latest -dofile {open_checkpoint_tcl}")
        os.chmod(open_checkpoint_script, 0o755)

        return True

    def run_conformal(self) -> bool:
        # Quit
        self.append("exit")

        # Write main dofile
        dofile = os.path.join(self.run_dir, f"{self.check}.tcl")
        self.write_contents_to_path("\n".join(self.output), dofile)

        # Build args
        args = self.start_cmd
        args.extend([
            "-nogui",
            "-color",
            "-tclmode",
            "-dofile", dofile
        ])
        if self.restart_checkpoint != "":
            args.extend([
                "-restart_checkpoint",
                self.restart_checkpoint
            ])

        # Temporarily disable colours/tag to make run output more readable.
        # TODO: think of a more elegant way to do this?
        HammerVLSILogging.enable_colour = False
        HammerVLSILogging.enable_tag = False
        self.run_executable(args, cwd=self.run_dir)
        # TODO: check for errors and deal with them
        # According to user guide:
        # Bit   Condition
        # 0     Internal error
        # 1     Exit status before comparison
        # 2     Command error
        # 3     Unmapped points
        # 4     Non-equivalent points
        # 5     Abort or uncompared points exist during any comparison
        # 6     Abort or uncompared points exist during last comparison
        HammerVLSILogging.enable_colour = True
        HammerVLSILogging.enable_tag = True

        # TODO: check that formal run was successful

        return True

tool = Conformal



#  asap7 plugin for Hammer.
#
#  See LICENSE for licence details.

import sys
import re
import os
import shutil
import glob
import subprocess
import textwrap
from types import new_class
from typing import NamedTuple, List, Optional, Tuple, Dict, Set, Any

from hammer.tech import HammerTechnology
from hammer.vlsi import HammerTool, HammerPlaceAndRouteTool, HammerDRCTool, MentorCalibreTool, TCLTool, HammerToolHookAction

class ASAP7Tech(HammerTechnology):
    """
    Override the HammerTechnology used in `hammer_tech.py`
    This class is loaded by function `load_from_json`, and will pass the `try` in `importlib`.
    """
    def post_install_script(self) -> None:
        try:
            # Prioritize gdstk
            self.gds_tool = __import__('gdstk')
        except ImportError:
            self.logger.info("gdstk not found, falling back to gdspy...")
            try:
                self.gds_tool = __import__('gdspy')
                assert('1.4' in self.gds_tool.__version__)
            except (ImportError, AssertionError):
                self.logger.error("Check your gdspy v1.4 installation! Unable to hack ASAP7 PDK.")
                shutil.rmtree(self.cache_dir)
                sys.exit()
        self.generate_multi_vt_gds()
        self.fix_icg_libs()

    def generate_multi_vt_gds(self) -> None:
        """
        PDK GDS only contains RVT cells.
        This patch will generate the other 3(LVT, SLVT, SRAM) VT GDS files.
        """
        try:
            os.makedirs(os.path.join(self.cache_dir, "GDS"))
        except:
            self.logger.info("Multi-VT GDS's already created")
            return None

        try:
            self.logger.info("Generating GDS for Multi-VT cells using {}...".format(self.gds_tool.__name__))

            stdcell_dir = self.get_setting("technology.asap7.stdcell_install_dir")
            orig_gds = os.path.join(stdcell_dir, "GDS/asap7sc7p5t_27_R_201211.gds")

            if self.gds_tool.__name__ == 'gdstk':
                # load original GDS
                asap7_original_gds = self.gds_tool.read_gds(infile=orig_gds)
                original_cells = asap7_original_gds.cells
                cell_list = list(map(lambda c: c.name, original_cells))
                # required libs
                multi_libs = {
                    "L": {
                        "lib": self.gds_tool.Library(),
                        "mvt_layer": 98
                        },
                    "SL": {
                        "lib": self.gds_tool.Library(),
                        "mvt_layer": 97
                        },
                    "SRAM": {
                        "lib": self.gds_tool.Library(),
                        "mvt_layer": 110
                        },
                }
                # create new libs
                for vt, multi_lib in multi_libs.items():
                    multi_lib['lib'].name = asap7_original_gds.name.replace('R', vt)

                for cell in original_cells:
                    # extract polygon from layer 100(the boundary for cell)
                    boundary_polygon = next(filter(lambda p: p.layer==100, cell.polygons))
                    for vt, multi_lib in multi_libs.items():
                        new_cell_name = cell.name.rstrip('R') + vt
                        cell_list.append(new_cell_name)
                        mvt_layer = multi_lib['mvt_layer']
                        # copy boundary_polygon to mvt_layer to mark the this cell is a mvt cell.
                        mvt_polygon = self.gds_tool.Polygon(boundary_polygon.points, multi_lib['mvt_layer'], 0)
                        mvt_cell = cell.copy(name=new_cell_name, deep_copy=True).add(mvt_polygon)
                        # add mvt_cell to corresponding multi_lib
                        multi_lib['lib'].add(mvt_cell)

                for vt, multi_lib in multi_libs.items():
                    # write multi_lib
                    new_gds = os.path.basename(orig_gds).replace('R', vt)
                    multi_lib['lib'].write_gds(os.path.join(self.cache_dir, 'GDS', new_gds))

            elif self.gds_tool.__name__ == 'gdspy':
                # load original GDS
                asap7_original_gds = self.gds_tool.GdsLibrary().read_gds(infile=orig_gds, units='import')
                original_cells = asap7_original_gds.cell_dict
                cell_list = list(map(lambda c: c.name, original_cells.values()))
                # required libs
                multi_libs = {
                    "L": {
                        "lib": self.gds_tool.GdsLibrary(),
                        "mvt_layer": 98
                        },
                    "SL": {
                        "lib": self.gds_tool.GdsLibrary(),
                        "mvt_layer": 97
                        },
                    "SRAM": {
                        "lib": self.gds_tool.GdsLibrary(),
                        "mvt_layer": 110
                        },
                }
                # create new libs
                for vt, multi_lib in multi_libs.items():
                    multi_lib['lib'].name = asap7_original_gds.name.replace('R', vt)

                for cell in original_cells.values():
                    poly_dict = cell.get_polygons(by_spec=True)
                    # extract polygon from layer 100(the boundary for cell)
                    boundary_polygon = poly_dict[(100, 0)]
                    for vt, multi_lib in multi_libs.items():
                        new_cell_name = cell.name.rstrip('R') + vt
                        cell_list.append(new_cell_name)
                        mvt_layer = multi_lib['mvt_layer']
                        # copy boundary_polygon to mvt_layer to mark the this cell is a mvt cell.
                        mvt_polygon = self.gds_tool.PolygonSet(boundary_polygon, multi_lib['mvt_layer'], 0)
                        mvt_cell = cell.copy(name=new_cell_name, exclude_from_current=True, deep_copy=True).add(mvt_polygon)
                        # add mvt_cell to corresponding multi_lib
                        multi_lib['lib'].add(mvt_cell)

                for vt, multi_lib in multi_libs.items():
                    # write multi_lib
                    new_gds = os.path.basename(orig_gds).replace('R', vt)
                    multi_lib['lib'].write_gds(os.path.join(self.cache_dir, 'GDS', new_gds))

            # Write out cell list for scaling script
            with open(os.path.join(self.cache_dir, 'stdcells.txt'), 'w') as f:
                f.writelines('{}\n'.format(cell) for cell in cell_list)

        except:
            os.rmdir(os.path.join(self.cache_dir, "GDS"))
            self.logger.error("GDS patching failed! Check your gdstk, gdspy, and/or ASAP7 PDK installation.")
            sys.exit()

    def fix_icg_libs(self) -> None:
        """
        ICG cells are missing statetable.
        """
        try:
            os.makedirs(os.path.join(self.cache_dir, "LIB/NLDM"))
        except:
            self.logger.info("ICG LIBs already fixed")
            return None

        try:
            self.logger.info("Fixing ICG LIBs...")
            statetable_text = "\\n".join([
               '\    statetable ("CLK ENA SE", "IQ") {',
                '      table : "L L L : - : L ,  L L H : - : H , L H L : - : H , L H H : - : H , H - - : - : N ";',
                '    }'])
            gclk_func = "CLK & IQ"
            lib_dir = os.path.join(self.get_setting("technology.asap7.stdcell_install_dir"), "LIB/NLDM")
            old_libs = glob.glob(os.path.join(lib_dir, "*SEQ*"))
            new_libs = list(map(lambda l: os.path.join(self.cache_dir, "LIB/NLDM", os.path.basename(l)), old_libs))

            for olib, nlib in zip(old_libs, new_libs):
                # Use gzip and sed directly rather than gzip python module
                # Add the statetable to ICG cells
                # Change function to state_function for pin GCLK
                subprocess.call(["gzip -cd {olib} | sed '/ICGx*/a {stbl}' | sed '/CLK & IQ/s/function/state_function/g' | gzip > {nlib}".format(olib=olib, stbl=statetable_text, nlib=nlib)], shell=True)
        except:
            os.rmdir(os.path.join(self.cache_dir, "LIB/NLDM"))
            os.rmdir(os.path.join(self.cache_dir, "LIB"))
            self.logger.error("Failed to fix ICG LIBs. Check your ASAP7 installation!")
            sys.exit()

    def get_tech_par_hooks(self, tool_name: str) -> List[HammerToolHookAction]:
        hooks = {"innovus": [
            HammerTool.make_post_persistent_hook("init_design", asap7_innovus_settings),
            HammerTool.make_post_insertion_hook("floorplan_design", asap7_update_floorplan),
            HammerTool.make_post_insertion_hook("write_design", asap7_scale_final_gds)
            ]}
        return hooks.get(tool_name, [])

    def get_tech_drc_hooks(self, tool_name: str) -> List[HammerToolHookAction]:
        hooks = {"calibre": [
            HammerTool.make_replacement_hook("generate_drc_run_file", asap7_generate_drc_run_file)
            ]}
        return hooks.get(tool_name, [])

def asap7_innovus_settings(ht: HammerTool) -> bool:
    assert isinstance(ht, HammerPlaceAndRouteTool), "Innovus settings only for par"
    assert isinstance(ht, TCLTool), "innovus settings can only run on TCL tools"
    """Settings that may need to be reapplied at every tool invocation
    Note that the particular routing layer settings here will persist in Innovus;
    this hook only serves as an example of what commands may need to persist."""
    ht.append('''
set_db route_design_bottom_routing_layer 2
set_db route_design_top_routing_layer 7

# Ignore 1e+31 removal arcs for ASYNC DFF cells
set_db timing_analysis_async_checks no_async

# Via preferences for stripes
set_db generate_special_via_rule_preference { M7_M6widePWR1p152 M6_M5widePWR1p152 M5_M4widePWR0p864 M4_M3widePWR0p864 M3_M2widePWR0p936 }

# Prevent extending M1 pins in cells
set_db route_design_with_via_in_pin true
    ''')
    return True

def asap7_update_floorplan(ht: HammerTool) -> bool:
    assert isinstance(ht, HammerPlaceAndRouteTool), "asap7_update_floorplan can only run on par"
    assert isinstance(ht, TCLTool), "asap7_update_floorplan can only run on TCL tools"
    """
    This is needed to block top/bottom site rows and re-do wiring tracks.
    This resolves many DRCs and removes the need for the user to do it in placement constraints.
    """
    ht.append('''
# Need to delete and recreate tracks based on tech LEF pitches but overriding offsets
add_tracks -honor_pitch -offsets { M4 horiz 0.048 M5 vert 0.048 M6 horiz 0.064 M7 vert 0.064 }

# Create place blockage on top & bottom row, fixes wiring issue + power vias for DRC/LVS
set core_lly [get_db current_design .core_bbox.ll.y]
set core_ury [expr [get_db current_design .core_bbox.ur.y] - 1.08]
set botrow [get_db rows -if {.rect.ll.y == $core_lly}]
set toprow [get_db rows -if {.rect.ur.y > $core_ury}]
create_place_blockage -area [get_db $botrow .rect] -name ROW_BLOCK_BOT
create_place_blockage -area [get_db $toprow .rect] -name ROW_BLOCK_TOP
''')
    return True

def asap7_scale_final_gds(ht: HammerTool) -> bool:
    assert isinstance(ht, HammerPlaceAndRouteTool), "asap7_scale_final_gds can only run on par"
    assert isinstance(ht, TCLTool), "asap7_scale_final_gds can only run on TCL tools"
    """
    Scale the final GDS by a factor of 4
    scale_gds_script writes the actual Python script to execute from the Tcl interpreter
    """
    # This is from the Cadence tools
    cadence_output_gds_name = ht.output_gds_filename  # type: ignore

    ht.append('''
# Innovus <19.1 appends some bad LD_LIBRARY_PATHS, so remove them before executing python
set env(LD_LIBRARY_PATH) [join [lsearch -not -all -inline [split $env(LD_LIBRARY_PATH) ":"] "*INNOVUS*"] ":"]
asap7_gds_scale {stdcells_file} {gds_file}
'''.format(stdcells_file = os.path.join(ht.technology.cache_dir, 'stdcells.txt'),
           gds_file = cadence_output_gds_name))
    return True

def asap7_generate_drc_run_file(ht: HammerTool) -> bool:
    assert isinstance(ht, HammerDRCTool), "asap7_generate_drc_run_file can only run on drc"
    assert isinstance(ht, MentorCalibreTool), "asap7_generate_drc_run_file can only run on a Calibre tool"
    """
    Replace drc_run_file to prevent conflicting SVRF statements
    Symlink DRC GDS to test.gds as expected by deck and results directories
    """
    # These are from the Calibre tools
    ht_drc_run_file = ht.drc_run_file  # type: ignore
    ht_max_drc_results = ht.max_drc_results  # type: ignore
    ht_virtual_connect_colon = ht.virtual_connect_colon  # type: ignore

    new_layout_file = os.path.join(ht.run_dir, 'test.gds')
    if not os.path.lexists(new_layout_file):
        os.symlink(os.path.splitext(ht.layout_file)[0] + '_drc.gds', new_layout_file)
    ht.layout_file = new_layout_file

    with open(ht_drc_run_file, "w") as f:
        f.write(textwrap.dedent("""
        // Generated by HAMMER

        DRC MAXIMUM RESULTS {max_results}
        DRC MAXIMUM VERTEX 4096

        DRC CELL NAME YES CELL SPACE XFORM

        VIRTUAL CONNECT COLON {virtual_connect}
        VIRTUAL CONNECT REPORT NO
        """).format(
            max_results=ht_max_drc_results,
            virtual_connect="YES" if ht_virtual_connect_colon else "NO"
        )
        )
        # Include paths to all supplied decks
        for rule in ht.get_drc_decks():
            f.write("INCLUDE \"{}\"\n".format(rule.path))
        # Note that an empty list means run all, and Calibre conveniently will do just that
        # if we don't specify any individual checks to run.
        if len(ht.drc_rules_to_run()) > 0:
            f.write("\nDRC SELECT CHECK\n")
        for check in ht.drc_rules_to_run():
            f.write("\t\"{}\"\n".format(check))
        f.write("\nDRC ICSTATION YES\n")
        f.write(ht.get_additional_drc_text())
    return True

tech = ASAP7Tech()



#  Mock hammer-vlsi sim plugin to help test Hammer infrastructure
#  without proprietary/NDAed tools.
#  NOT FOR EXTERNAL/PUBLIC USE.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerSimTool, DummyHammerTool, HammerToolStep, deepdict
from hammer.config import HammerJSONEncoder

from typing import Dict, List, Any, Optional
from decimal import Decimal

import os
import json


class MockSim(HammerSimTool, DummyHammerTool):

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = deepdict(super().env_vars)
        new_dict.update({})  # TODO: stuffs
        return new_dict

    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.step1
        ])

    @property
    def force_regs_file_path(self) -> str:
        return os.path.join(self.run_dir, "{module}_regs.txt".format(module=self.top_module))

    def step1(self) -> bool:
        sim_file = os.path.join(self.run_dir, "{module}_regs.txt".format(module=self.top_module))
        with open(sim_file, "w") as f:
            f.write(self.get_setting("sim.inputs.level"))
            f.write("\n")
            for option in self.get_setting("sim.inputs.options"):
                f.write(option)
                f.write("\n")
        return True



tool = MockSim



#  get-config
#
#  Read a config from either the given JSON database (if present) or the HAMMER_DATABASE environment variable.
#
#  See LICENSE for licence details.

# pylint: disable=invalid-name

import argparse
import json
import os
import sys

import hammer.config as hammer_config

def run(args):
    if args.db is None:
        try:
            db_location = os.environ["HAMMER_DATABASE"]
        except KeyError:
            print("No database --db specified and HAMMER_DATABASE is not defined", file=sys.stderr)
            return 1
    else:
        db_location = args.db
    database = hammer_config.HammerDatabase()
    # TODO(edwardw): rethink this hack? This simply treats the entire exported JSON as a "project JSON", which might actually be ok.
    database.update_project([json.load(open(db_location))])
    try:
        print(str(database.get_setting(args.key, args.nullvalue)))
        return 0
    except ValueError as e:
        print("Error: " + e.args[0], file=sys.stderr)
        return 1

def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("-n", "--nullvalue", default="null", required=False,
                        help='Value to print out for nulls. (default: "null")')
    parser.add_argument("-e", "--error-if-missing", action='store_const',
                        const=True, default=False, required=False,
                        help="Error out if the key is missing. (default: false)")
    parser.add_argument('--db', type=str, required=False,
                        help='Path to the JSON database')
    parser.add_argument('key', metavar='KEY', type=str,
                        help='Key to retrieve from the database')

    sys.exit(run(parser.parse_args()))



# Netgen LVS plugin for Hammer
#
# See LICENSE for licence details.

import os
from textwrap import dedent as dd
from typing import List, Optional, Dict, Any

from hammer.logging import HammerVLSILogging
from hammer.utils import deepdict, get_or_else
from hammer.vlsi import HammerToolStep
from hammer.vlsi import HammerLVSTool, TCLTool
import hammer.tech as hammer_tech

class Netgen(HammerLVSTool, TCLTool):

    #=========================================================================
    # overrides from parent classes
    #=========================================================================
    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.run_ext2spice,
            self.run_lvs
        ])

    def do_post_steps(self) -> bool:
        assert super().do_post_steps()
        return self.run_netgen()

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = deepdict(super().env_vars)
        return new_dict

    def fill_outputs(self) -> bool:
        return True

    def tool_config_prefix(self) -> str:
        return "lvs.netgen"

    def erc_results_pre_waived(self) -> Dict[str, int]:
        return {}

    def globally_waived_erc_rules(self) -> List[str]:
        return []

    def lvs_results(self) -> List[str]:
        #TODO: extract LVS descriptions from netgen output
        return []

    def version_number(self, version:str) -> int:
        """Get version from netgen bin"""
        version = self.run_executable([self.get_setting("lvs.netgen.netgen_bin"), "-batch"])
        return int(version.replace(".", ""))

    @property
    def ext2spice_netlist(self) -> str:
        return os.path.join(self.run_dir, self.top_module + ".ext.sp")

    @property
    def generated_scripts_dir(self) -> str:
        return os.path.join(self.run_dir, "generated-scripts")

    @property
    def view_lvs_script(self) -> str:
        return os.path.join(self.generated_scripts_dir, "view_lvs")

    #=========================================================================
    # useful subroutines
    #=========================================================================

    def run_netgen(self) -> bool:
        run_script = os.path.join(self.run_dir, "lvs.tcl")
        with open(run_script, "w") as f:
            f.write("\n".join(self.output))

        """
        Create view_lvs script. This opens interactive window but has to run LVS
        all over again because there is no LVS database that can be loaded in.
        TODO: parse the JSON output too, like OpenLane
        """
        os.makedirs(self.generated_scripts_dir, exist_ok=True)
        with open(self.view_lvs_script, "w") as f:
            f.write(dd("""
            cd {run_dir}
            {netgen_bin} source {run_script}
            """.format(
                run_dir = self.run_dir,
                netgen_bin = self.get_setting("lvs.netgen.netgen_bin"),
                run_script = run_script
                )))
        os.chmod(self.view_lvs_script, 0o755)

        args = [self.get_setting("lvs.netgen.netgen_bin"), "-batch", "source", run_script]


        # Finally append the no GUI options and full Tcl run script
        if bool(self.get_setting("lvs.netgen.generate_only")):
            self.logger.info("Generate-only mode: command-line is " + \
                             " ".join(args))
        else:
            # Temporarily disable colors/tag to make run output more readable
            # TODO: think of a more elegant way to do this?
            HammerVLSILogging.enable_colour = False
            HammerVLSILogging.enable_tag = False
            self.run_executable(args, cwd=self.run_dir) # TODO: check for errors
            HammerVLSILogging.enable_colour = True
            HammerVLSILogging.enable_tag = True

        return True

    #========================================================================
    # LVS main steps
    #========================================================================
    def run_ext2spice(self) -> bool:
        """
        Run magic ext2spice
        Adapted from OpenLane
        """
        # Skip if generated netlist newer than layout file
        if os.path.exists(self.ext2spice_netlist):
            if os.path.getmtime(self.layout_file) < os.path.getmtime(self.ext2spice_netlist):
                self.logger.info("Layout Spice file already generated, skipping...")
                return True

        # Magic args
        args = [self.get_setting("drc.magic.magic_bin"), "-noconsole", "-dnull"]
        rcfile = self.get_setting("drc.magic.rcfile")
        # Equivalent to get_drc_decks() for DRCTool
        techfile = self.technology.get_drc_decks_for_tool(self.get_setting("vlsi.core.drc_tool"))

        if rcfile is not None:
            args.extend(["-rcfile", rcfile])
        else:
            # DRC deck should be the tech file. There should only be 1.
            if len(techfile) == 0 or len(techfile) > 1:
                self.logger.error("None or more than 1 tech file (DRC deck) found. netgen only supports 1.")
            args.extend(["-T", techfile[0]])

        ext2spice_script = os.path.join(self.run_dir, "ext2spice.tcl")
        args.append(ext2spice_script)

        # Create folder for all .ext files
        os.makedirs(os.path.join(self.run_dir, "ext2spice-rundir"), exist_ok=True)

        # ext2spice script
        self.append("gds read " + self.layout_file)
        self.append("load " + self.top_module)
        self.append("cd ext2spice-rundir")
        self.append("extract do local") # extract into ext2spice-rundir
        self.append("extract no capacitance")
        self.append("extract no coupling")
        self.append("extract no resistance")
        self.append("extract no adjust")
        if not self.get_setting("lvs.netgen.connect_by_label"):
            self.append("extract unique")
        self.append("extract warn all")
        self.append("extract")
        self.append("ext2spice lvs")
        self.append("ext2spice -o " + self.ext2spice_netlist)
        self.append("feedback save ext2spice.log")
        self.append("quit")

        with open(ext2spice_script, "w") as f:
            f.write("\n".join(self.output))
        self.output.clear()

        if bool(self.get_setting("lvs.netgen.generate_only")):
            self.logger.info("Generate-only mode: command-line is " + \
                             " ".join(args))
        else:
            # Temporarily disable colors/tag to make run output more readable
            # TODO: think of a more elegant way to do this?
            HammerVLSILogging.enable_colour = False
            HammerVLSILogging.enable_tag = False
            self.run_executable(args, cwd=self.run_dir) # TODO: check for errors
            HammerVLSILogging.enable_colour = True
            HammerVLSILogging.enable_tag = True

        return True

    def run_lvs(self) -> bool:
        """
        Read all libraries (stdcells, macros), then run comparison
        Additional LVS text can then modify netlists
        TODO: break up lvs command into constituent parts for more control
        """
        # libs for circuit1. spice format must be specified.
        self.append('puts "Reading layout..."')
        self.append("readnet spice {} 1".format(self.ext2spice_netlist))

        # libs for circuit2. auto format detection (somewhat dangerous).
        self.append('puts "Reading schematics..."')
        library_spice_files = self.technology.read_libs([hammer_tech.filters.spice_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
        ilms = list(map(lambda x: x.netlist, self.ilms))  # type: List[str]
        for sch in self.schematic_files + ilms + library_spice_files:
            self.append('puts "Reading {}..."'.format(sch))
            self.append("readnet {} 2".format(sch))

        self.append(self.get_additional_lvs_text())

        self.append('puts "Running LVS..."')
        setup_file = self.get_setting("lvs.netgen.setup_file")
        self.append("lvs {{{top} 1}} {{{top} 2}} {setup} {top}.lvs.log -json".format(
                    top=self.top_module,
                    setup=get_or_else(setup_file, "nosetup")))
        return True

tool = Netgen



#!/usr/bin/env python3
# type: ignore
#   tell mypy to ignore this file during typechecking
# -*- coding: utf-8 -*-
#
#  Script to generate the SRAM22 sram-cache.json
#
#  See LICENSE for licence details.

import sys
import re
import json

from typing import List

def main(args: List[str]) -> int:
    if len(args) != 3:
        print("Usage: ./sram-cache-gen.py list-of-srams-1-per-line.txt output-file.json")
        print("E.g.: ./sram-cache-gen.py srams.txt sram-cache.json")
        return 1

    list_of_srams = []  # type: List[str]
    with open(sys.argv[1]) as f:
        for line in f:
            list_of_srams.append(line)

    print(str(len(list_of_srams)) + " SRAMs to cache")

    sram_dicts = []

    for sram_name in list_of_srams:
        # SRAM22-generated single-port RAMs
        if sram_name.startswith("sramgen_"):
            match = re.match(r"sramgen_sram_(\d+)x(\d+)m(\d+)w(\d+)(\D*)", sram_name)
            if match:
                width = int(match.group(2))
                mask_gran = int(match.group(4))
                
                sram_dict = {}
                sram_dict['type'] = 'sram'
                sram_dict['name'] = sram_name.strip()
                sram_dict['source'] = 'sram22'
                sram_dict['depth'] = match.group(1)
                sram_dict['width'] = width
                sram_dict['family'] = '1rw'
                sram_dict['mask'] = "true"
                sram_dict['vt'] = 'svt'
                sram_dict['mux'] = int(match.group(3))
                sram_dict['ports'] = []

                port_dict = {}
                port_dict['address port name'] = "addr"
                port_dict['address port polarity'] = "active high"
                
                port_dict['clock port name'] = "clk"
                port_dict['clock port polarity'] = "active high"

                port_dict['write enable port name'] = "we"
                port_dict['write enable port polarity'] = "active high" # ???

                port_dict['output port name'] = "dout"
                port_dict['output port polarity'] = "active high"

                port_dict['input port name'] = "din"
                port_dict['input port polarity'] = "active high"
                
                if mask_gran != width:
                    port_dict['mask port name'] = "wmask"
                    port_dict['mask granularity'] = mask_gran
                    port_dict['mask port polarity'] = "active high"  # ???
                
                sram_dict['ports'].append(port_dict.copy())

                sram_dicts.append(sram_dict.copy())
            
            else:
                print("Unsupported memory: {n}".format(n=sram_name), file=sys.stderr)
                return 1
        else:
            print("Unsupported memory: {n}".format(n=sram_name), file=sys.stderr)
            return 1

    with open(sys.argv[2], "w") as f:
        json.dump(sram_dicts, f, indent=2)

    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))



#  Data structures to represent technology stackup options.
#
#  See LICENSE for licence details.

from decimal import Decimal
from enum import Enum
from functools import partial
from typing import Any, List, Tuple, Optional

from pydantic import BaseModel, root_validator

from hammer.utils import coerce_to_grid
from hammer.logging import HammerVLSILoggingContext


class RoutingDirection(str, Enum):
    """
    Represents a preferred routing direction for a metal layer.
    Note that this represents a *preferred* direction, not a DRC rule.
    """
    Vertical = "vertical"
    Horizontal = "horizontal"
    Redistribution = "redistribution"

    def opposite(self) -> "RoutingDirection":
        """
        Return the opposite routing direction.
        For Redistribution, this returns itself.
        :return: Opposite routing direction
        """
        if self == RoutingDirection.Vertical:
            return RoutingDirection.Horizontal
        elif self == RoutingDirection.Horizontal:
            return RoutingDirection.Vertical
        else:
            return self


class WidthSpacingTuple(BaseModel):
    """
    A tuple of wire width limit and spacing for generating a piecewise linear rule
    for spacing based on wire width.

    width_at_least: Any wires larger than this must obey the minSpacing rule.
    min_spacing: The minimum spacing for this bin.
                 If a wire is wider than multiple entries, the worst-case (larger)
                 minSpacing wins.
    """
    width_at_least: Decimal
    min_spacing: Decimal

    @staticmethod
    def from_setting(grid_unit: Decimal, d: dict) -> "WidthSpacingTuple":
        width_at_least = coerce_to_grid(d["width_at_least"], grid_unit)
        min_spacing = coerce_to_grid(d["min_spacing"], grid_unit)
        assert width_at_least >= 0
        assert min_spacing > 0
        return WidthSpacingTuple(
            width_at_least=width_at_least,
            min_spacing=min_spacing
        )

    @staticmethod
    def from_list(grid_unit: Decimal, l: List[dict]) -> List["WidthSpacingTuple"]:
        out = sorted(list(map(partial(WidthSpacingTuple.from_setting, grid_unit), l)), key=lambda x: x.width_at_least)

        # Check that spacings increase.
        current_spacing = Decimal(0)
        for wst in out:
            assert wst.min_spacing >= current_spacing
            current_spacing = wst.min_spacing
        return out


class Metal(BaseModel):
    """
    A metal layer and some basic info about it.

    name: Metal layer name (e.g. M1, M2).
    index: The order in the stackup (lower is closer to the substrate).
    direction: The preferred routing direction of this metal layer, or
               RoutingDirection.Redistribution for non-routing top-level
               redistribution metals like Aluminium.
    min_width: The minimum wire width for this layer.
    max_width: The maximum wire width for this layer.
    pitch: The minimum cross-mask pitch for this layer (NOT same-mask pitch
           for multiple-patterned layers). Width of routing grid for a given layer. 
           To route denser wires on chip, multiple masks are required. 
           During fabrication, the masks are applied separately with some spatial offsets 
           to achieve denser line patterning. For more information on multiple-patterning, 
           check https://en.wikipedia.org/wiki/Multiple_patterning
    offset: The routing track offset from the origin for the first track in this layer.
            (0 = first track is on an axis).
    power_strap_widths_and_spacings: A list of WidthSpacingTuples that specify the minimum
                                     spacing rules for an infinitely long wire of variying width.
    power_strap_width_table: A list of allowed metal widths in the technology.
                             Widths smaller than the last number must be quantized to a value in the table.
    grid_unit: The fixed-point decimal value of a minimum grid unit (e.g. 1nm = 0.001).
               For most technologies, this comes from the technology plugin and is the same for all layers.
    """
    name: str
    index: int
    direction: RoutingDirection
    min_width: Decimal
    max_width: Optional[Decimal] = None
    pitch: Decimal
    offset: Decimal
    power_strap_widths_and_spacings: List[WidthSpacingTuple]
    power_strap_width_table: List[Decimal] = []
    # Note: grid_unit is not currently parsed as part of the Metal data structure!
    # See #379
    grid_unit: Decimal

    class Config:
        use_enum_values = True

    @root_validator(pre=True)
    def widths_must_snap_to_grid(cls, values):
        grid_unit = Decimal(str(values.get("grid_unit")))
        for field in ["min_width", "pitch", "offset"]:
            raw_value = Decimal(str(values.get(field)))
            snapped_value = coerce_to_grid(raw_value, grid_unit)
            if raw_value != snapped_value:
                raise ValueError(f"{field} ({raw_value}) is not aligned to the grid_unit ({grid_unit})")
        if values.get("power_strap_width_table"):
            for width in values.get("power_strap_width_table"):
                width = Decimal(str(width))
                snapped_width = coerce_to_grid(width, grid_unit)
                if width != snapped_width:
                    raise ValueError(f"Width {width} in power_strap_width_table is not aligned to the grid_unit ({grid_unit})")
        if values.get("max_width") is not None:
            max_width = Decimal(str(values.get("max_width")))
            snapped_width = coerce_to_grid(max_width, grid_unit)
            if max_width != snapped_width:
                raise ValueError(f"max_width {max_width} is not aligned to the grid unit ({grid_unit})")
        return values

    @staticmethod
    def from_setting(grid_unit: Decimal, d: dict) -> "Metal":
        """
        Return a Metal object from a dict with keys "name", "index", "direction", "min_width", "max_width", "pitch", "offset", "power_strap_widths_and_spacings", and "power_strap_width_table"

        :param grid_unit: The manufacturing grid unit in um
        :param d: A dict containing the keys "name", "index", "direction", "min_width", "max_width", "pitch", "offset", "power_strap_widths_and_spacings", and "power_strap_width_table"
        """
        return Metal(
            grid_unit=grid_unit,
            name=str(d["name"]),
            index=int(d["index"]),
            direction=RoutingDirection(d["direction"]),
            min_width=coerce_to_grid(d["min_width"], grid_unit),
            max_width=coerce_to_grid(d["max_width"], grid_unit) if "max_width" in d and d["max_width"] is not None else None,
            pitch=coerce_to_grid(d["pitch"], grid_unit),
            offset=coerce_to_grid(d["offset"], grid_unit),
            power_strap_widths_and_spacings=WidthSpacingTuple.from_list(grid_unit, d["power_strap_widths_and_spacings"]),
            power_strap_width_table=Metal.power_strap_widths_from_list(grid_unit, d["power_strap_width_table"] if "power_strap_width_table" in d and d["power_strap_width_table"] else [])
        )

    @staticmethod
    def power_strap_widths_from_list(grid_unit: Decimal, l: List[Any]) -> List[Decimal]:
        """
        Read and cocerce wire widths from the technology LEF width table.
        """
        return sorted(map(lambda w: coerce_to_grid(w, grid_unit), map(float, l)))

    def get_spacing_for_width(self, width: Decimal) -> Decimal:
        """
        Get the minimum spacing for a provided width.

        :param width: Width to calculate minimum spacing for.
        :return: Minimum spacing for `width`
        """
        spacing = Decimal(0)
        for wst in self.power_strap_widths_and_spacings:
            if width >= wst.width_at_least:
                spacing = max(spacing, wst.min_spacing)
            else:
                # The list is sorted so we can early-out
                return spacing
        return spacing

    def min_spacing_and_max_width_from_pitch(self, pitch: Decimal) -> Tuple[Decimal, Decimal]:
        """
        Derive the minimum spacing and maximally-sized wire for a
        desired pitch.

        Use this when the wire width is unknown, but you know the pitch.
        This calculation essentially plots the wire width on the X axis
        and the minimum pitch on the Y axis.

        You'll see discontinuites at the width-spacing table entries.
        If the desired pitch falls on a sloped line (i.e. > min width for
        entry N but less than min width for entry N+1), pick that spacing.
        If the desired pitch falls on a vertical line, pick the maximum width
        entry for N, which is the entry for N+1 minus delta (2 grid units),
        and then the spacing will be larger than the min spacing.

        :param pitch: Desired pitch
        :return: Tuple of (minimum spacing, maximally-sized wire)
        """
        widths_and_spacings = self.power_strap_widths_and_spacings
        spacing = widths_and_spacings[0].min_spacing
        for first, second in zip(widths_and_spacings[:-1], widths_and_spacings[1:]):
            if pitch >= (second.min_spacing + second.width_at_least):
                spacing = second.min_spacing
            elif pitch >= (first.min_spacing + second.width_at_least):
                # we are asking for a pitch that is width-constrained
                width = second.width_at_least - (self.grid_unit*2)
                spacing = pitch - width

        width = pitch - spacing
        if self.max_width and self.max_width > 0.0 and width > self.max_width:
            width = self.max_width
        if width < 0:
            raise ValueError("Desired pitch {pitch} is illegal".format(pitch=pitch))
        return spacing, pitch - spacing

    def min_spacing_from_pitch(self, pitch: Decimal) -> Decimal:
        """
        Derive the minimum spacing for a maximally-sized wire given a
        desired pitch.
        See min_spacing_and_max_width_from_pitch for details.

        :param pitch: Desired pitch
        :return: Minimum spacing for said pitch.
        """
        return self.min_spacing_and_max_width_from_pitch(pitch)[0]

    def max_width_from_pitch(self, pitch: Decimal) -> Decimal:
        """
        Derive the maximum wire width for a maximally-sized wire given a
        desired pitch.
        See min_spacing_and_max_width_from_pitch for details.

        :param pitch: Desired pitch
        :return: Maximum wire width for said pitch.
        """
        return self.min_spacing_and_max_width_from_pitch(pitch)[1]

    def quantize_to_width_table(self, width: Decimal, layer: str, logger: Optional[HammerVLSILoggingContext]) -> Decimal:
        """
        Compare a desired width to the width table, if specified for a technology.
        Will return the allowable width smaller than or equal to the desired width,
        except if the desired width is greater than or equal to the last width in the width table.
        Issues a logger warning for the user if the returned width was quantized.
        """
        width_table = self.power_strap_width_table
        if len(width_table) == 0:
            qwidth = width
        else:
            for i, w in enumerate(width_table):
                if width > w:
                    # The last entry in table is special: width can be greater
                    # than or equal to this number.
                    if i == len(width_table) - 1:
                        qwidth = width
                        break
                    else:
                        continue
                elif width == w:
                    qwidth = w
                    break
                else:
                    if logger is not None:
                        logger.warning("The desired power strap width {dw} on {lay} was quantized down to {fw} based on the technology's width table. Please check your power grid.".format(dw=str(width), lay=layer, fw=str(width_table[i-1])))
                    qwidth = width_table[i-1]
                    break
        return qwidth

    def get_width_spacing_start_twt(self,
                                    tracks: int,
                                    logger: Optional[HammerVLSILoggingContext]
                                   ) -> Tuple[Decimal, Decimal, Decimal]:
        """
        This method will return the maximum width a wire can be in order
        to consume a given number of routing tracks.
        This assumes the neighbors of the wide wire are minimum-width routes.
        i.e. T W T
        T = thin / min-width
        W = wide
        See min_spacing_and_max_width_from_pitch for an explanation of the calculation.

        :param tracks: Number of routing tracks to consume
        :return: Returns tuple of (width, spacing, start)
        """
        widths_and_spacings = self.power_strap_widths_and_spacings
        spacing = widths_and_spacings[0].min_spacing
        # the T W T pattern contains one wires (W) and 2 spaces (S2)
        s2w = (tracks + 1) * self.pitch - self.min_width

        assert int(s2w / self.grid_unit) % 2 == 0, "This calculation should always produce an even s2w"

        width = s2w - spacing*2
        for first, second in zip(widths_and_spacings[:-1], widths_and_spacings[1:]):
            if s2w >= second.min_spacing * 2 + second.width_at_least:
                spacing = second.min_spacing
                width = s2w - spacing * 2
            elif s2w >= first.min_spacing * 2 + second.width_at_least:
                # we are asking for a pitch that is width-constrained
                if int(second.width_at_least / self.grid_unit) % 2 == 0:
                    # even
                    width = second.width_at_least - (self.grid_unit * 2)
                else:
                    # odd
                    width = second.width_at_least - self.grid_unit
                spacing = (s2w - width) / 2
        if self.max_width and self.max_width > 0.0 and width > self.max_width:
            width = self.max_width
            spacing = (s2w - width) / 2

        assert int(self.min_width / self.grid_unit) % 2 == 0, (
            "Assuming all min widths are even here, if not fix me")
        assert int(width / self.grid_unit) % 2 == 0, (
            "This calculation should always produce an even width")
        start = self.min_width / 2 + spacing
        return (self.quantize_to_width_table(width, self.name, logger), spacing, start)

    def get_width_spacing_start_twwt(self, tracks: int, logger: Optional[HammerVLSILoggingContext], force_even: bool = False) -> Tuple[Decimal, Decimal, Decimal]:
        """
        This method will return the maximum width a wire can be in order
        to consume a given number of routing tracks.
        This assumes that the wires are in the following configuration.
        i.e. T W W T
        T = thin / min-width
        W = wide
        See min_spacing_and_max_width_from_pitch for an explanation of the calculation.

        :param tracks: Number of routing tracks to consume
        :param force_even: Forces the width of the wire to be an even multiple of the unit grid
        :return: Returns tuple of (width, spacing, start)
        """
        widths_and_spacings = self.power_strap_widths_and_spacings
        spacing = widths_and_spacings[0].min_spacing
        assert self.pitch - self.min_width == spacing, "Tech plugin is malformed for metal {}, the minimum spacing in the width-spacing list must be the same as (pitch - min_width).".format(self.name)
        # the T W W T pattern contains two wires (W2) and 3 spaces (S3)
        s3w2 = ((2 * tracks) + 1) * self.pitch - self.min_width
        width = (s3w2 - spacing * 3) / 2
        for first, second in zip(widths_and_spacings[:-1], widths_and_spacings[1:]):
            if s3w2 >= second.min_spacing * 3 + second.width_at_least * 2:
                spacing = second.min_spacing
                width = (s3w2 - spacing * 3) / 2
            elif s3w2 >= first.min_spacing * 3 + second.width_at_least * 2:
                # we are asking for a pitch that is width-constrained
                width = second.width_at_least - (self.grid_unit * 1)
                spacing = (s3w2 - width * 2) / 3
        if self.max_width and self.max_width > 0.0 and width > self.max_width:
            # If we cannot maximize width, set it to max_width
            width = self.max_width
            spacing = (s3w2 - width * 2) / 3

        assert int(self.min_width / self.grid_unit) % 2 == 0, "Assuming all min widths are even here, if not fix me"
        start = self.min_width / 2 + spacing
        if force_even and int(width / self.grid_unit) % 2 == 1:
            width = width - self.grid_unit
            start = start + self.grid_unit
        return (self.quantize_to_width_table(width, self.name, logger), spacing, start)

    # TODO implement M W X* W M style wires, where X is slightly narrower
    # than W and centered on-grid.


class Stackup(BaseModel):
    """
    A stackup is a list of metals with a meaningful keyword name (for now).

    TODO: add vias, etc when we need them
    """
    grid_unit: Decimal
    name: str
    metals: List[Metal]

    @staticmethod
    def from_setting(grid_unit: Decimal, d: dict) -> "Stackup":
        # pylint: disable=missing-docstring
        return Stackup(
            grid_unit=grid_unit,
            name=str(d["name"]),
            metals=list(map(lambda x: Metal.from_setting(grid_unit, x), list(d["metals"])))
        )

    def get_metal(self, name: str) -> Metal:
        """
        Get a given metal layer by name.

        :param name: Name of the metal layer
        :return: A metal layer object
        """
        for m in self.metals:
            if m.name == name:
                return m
        raise ValueError("Metal named %s is not defined in stackup %s" % (name, self.name))

    def get_metal_by_index(self, index: int) -> Metal:
        """
        Get a given metal layer by index.

        :param index: Index of the metal layer
        :return: A metal layer object
        """
        for m in self.metals:
            if m.index == index:
                return m
        raise ValueError("Metal with index %d is not defined in stackup %s" % (index, self.name))



#  nop.py
#  No-op LVS tool.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerLVSTool, DummyHammerTool
from typing import List, Dict


class NopLVS(HammerLVSTool, DummyHammerTool):
    def fill_outputs(self) -> bool:
        return True

    def globally_waived_erc_rules(self) -> List[str]:
        return []

    def erc_results_pre_waived(self) -> Dict[str, int]:
        return {}

    def lvs_results(self) -> List[str]:
        return []


tool = NopLVS



#  hammer-vlsi plugin for Cadence Voltus.
#
#  See LICENSE for licence details.

from typing import List, Dict, Optional, Callable

import os
import json

from hammer.config import HammerJSONEncoder
from hammer.utils import in_place_unique
from hammer.vlsi import HammerPowerTool, HammerToolStep, MMMCCorner, MMMCCornerType, TimeValue, VoltageValue, FlowLevel
from hammer.logging import HammerVLSILogging
import hammer.tech as hammer_tech
from hammer.tech.specialcells import CellType
from hammer.common.cadence import CadenceTool


class Voltus(HammerPowerTool, CadenceTool):
    @property
    def post_synth_sdc(self) -> Optional[str]:
        # No post-synth SDC input for power...
        return None

    def tool_config_prefix(self) -> str:
        return "power.voltus"

    @property
    def env_vars(self) -> Dict[str, str]:
        new_dict = dict(super().env_vars)
        new_dict["VOLTUS_BIN"] = self.get_setting("power.voltus.voltus_bin")
        return new_dict

    @property
    def extra_corners_only(self) -> bool:
        return self.get_setting("power.inputs.extra_corners_only")

    @property
    def tech_stdcell_pgv_tcl(self) -> str:
        return os.path.join(self.run_dir, "tech_stdcell_pgv.tcl")

    @property
    def macro_pgv_tcl(self) -> str:
        return os.path.join(self.run_dir, "macro_pgv.tcl")

    @property
    def tech_lib_dir(self) -> str:
        return os.path.join(self.technology.cache_dir, "tech_pgv")

    @property
    def stdcell_lib_dir(self) -> str:
        return os.path.join(self.technology.cache_dir, "stdcell_pgv")

    @property
    def macro_lib_dir(self) -> str:
        return os.path.join(self.technology.cache_dir, "macro_pgv")

    @property
    def ran_tech_stdcell_pgv(self) -> bool:
        """ init_technology sets this to True if tech/stdcell PG views were generated by Hammer or provided by PDK """
        return self.attr_getter("_ran_tech_stdcell_pgv", False)

    @ran_tech_stdcell_pgv.setter
    def ran_tech_stdcell_pgv(self, val: bool) -> None:
        self.attr_setter("_ran_tech_stdcell_pgv", val)

    @property
    def gen_tech_stdcell_pgv(self) -> bool:
        """ init_technology sets this to True if stdcell PG views need to be generated """
        return self.attr_getter("_gen_tech_stdcell_pgv", False)

    @gen_tech_stdcell_pgv.setter
    def gen_tech_stdcell_pgv(self, val: bool) -> None:
        self.attr_setter("_gen_tech_stdcell_pgv", val)

    @property
    def ran_macro_pgv(self) -> bool:
        """ init_technology sets this to True if macro PG views were generated by Hammer """
        return self.attr_getter("_ran_macro_pgv", False)

    @ran_macro_pgv.setter
    def ran_macro_pgv(self, val: bool) -> None:
        self.attr_setter("_ran_macro_pgv", val)

    @property
    def gen_macro_pgv(self) -> bool:
        """ init_technology sets this to True if macro PG views need to be generated """
        return self.attr_getter("_gen_macro_pgv", False)

    @gen_macro_pgv.setter
    def gen_macro_pgv(self, val: bool) -> None:
        self.attr_setter("_gen_macro_pgv", val)

    @property
    def macro_pgv_cells(self) -> List[str]:
        """ init_technology populates the list of macros to generate PG views for """
        return self.attr_getter("_macro_pgv_cells", [])

    @macro_pgv_cells.setter
    def macro_pgv_cells(self, val: List[str]) -> None:
        self.attr_setter("_macro_pgv_cells", val)

    def tech_lib_filter(self) -> List[Callable[[hammer_tech.Library], bool]]:
        """ Filter only libraries from tech plugin """
        return [self.filter_for_tech_libs]

    def filter_for_tech_libs(self, lib: hammer_tech.Library) -> bool:
        return lib in self.technology.tech_defined_libraries

    def extra_lib_filter(self) -> List[Callable[[hammer_tech.Library], bool]]:
        """ Filter only libraries from vlsi.inputs.extra_libraries """
        return [self.filter_for_extra_libs]

    def filter_for_extra_libs(self, lib: hammer_tech.Library) -> bool:
        return lib in list(map(lambda el: el.store_into_library(), self.technology.get_extra_libraries()))

    def get_mmmc_pgv(self, corner: MMMCCorner) -> List[str]:
        return self.technology.read_libs([hammer_tech.filters.power_grid_library_filter],
                                             hammer_tech.HammerTechnologyUtils.to_plain_item,
                                             extra_pre_filters=[
                                                 self.filter_for_mmmc(voltage=corner.voltage, temp=corner.temp)])

    def get_mmmc_spice_models(self, corner: MMMCCorner) -> List[str]:
        return self.technology.read_libs([hammer_tech.filters.spice_model_file_filter],
                                             hammer_tech.HammerTechnologyUtils.to_plain_item,
                                             extra_pre_filters=[
                                                 self.filter_for_mmmc(voltage=corner.voltage, temp=corner.temp)])

    def get_mmmc_spice_corners(self, corner: MMMCCorner) -> List[str]:
        return self.technology.read_libs([hammer_tech.filters.spice_model_lib_corner_filter],
                                             hammer_tech.HammerTechnologyUtils.to_plain_item,
                                             extra_pre_filters=[
                                                 self.filter_for_mmmc(voltage=corner.voltage, temp=corner.temp)],
                                             must_exist=False)

    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.check_level,
            self.init_technology,
            self.init_design,
            self.static_power,
            self.active_power,
            self.static_rail,
            self.active_rail,
            self.run_voltus
        ])

    def check_level(self) -> bool:
        if self.level == FlowLevel.PAR:
            return True
        else:
            self.logger.error("The FlowLevel is invalid. The Voltus plugin only supports post-par analysis. Check your power tool setting and flow step.")
            return False


    def init_technology(self) -> bool:
        corners = self.get_mmmc_corners()

        # Options for set_pg_library_mode
        base_options = ["-enable_distributed_processing", "true"]  # type: List[str]
        if self.get_setting("power.voltus.lef_layer_map"):
            base_options.extend(["-lef_layer_map", self.get_setting("power.voltus.lef_layer_map")])

        # Setup commands for each PG library run
        base_cmds = ["set_db design_process_node {}".format(self.get_setting("vlsi.core.node"))]
        base_cmds.append("set_multi_cpu_usage -local_cpu {}".format(self.get_setting("vlsi.core.max_threads")))

        # First, check if tech plugin supplies power grid libraries
        tech_pg_libs = self.technology.read_libs([hammer_tech.filters.power_grid_library_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.tech_lib_filter())
        tech_lib_lefs = self.technology.read_libs([hammer_tech.filters.lef_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.tech_lib_filter())
        if len(tech_pg_libs) > 0:
            self.logger.info("Technology already provides PG libraries. Moving onto macro PG libraries.")
        # Else, characterize tech & stdcell libraries only once
        elif not os.path.isdir(self.tech_lib_dir) or not os.path.isdir(self.stdcell_lib_dir):
            self.logger.info("Generating techonly and stdcell PG libraries for the first time...")
            ts_output = base_cmds.copy()
            # Get only the tech-defined libraries
            ts_output.append("read_physical -lef {{ {} }}".format(" ".join(tech_lib_lefs)))

            tech_options = base_options.copy()
            tech_options.extend(["-cell_type", "techonly"])
            # fillers, decaps
            stdfillers = self.technology.get_special_cell_by_type(CellType.StdFiller)
            decaps = self.technology.get_special_cell_by_type(CellType.Decap)
            tech_lib_sp = self.technology.read_libs([hammer_tech.filters.spice_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.tech_lib_filter())

            if not corners:
                # Start with tech-only library
                options = tech_options.copy()
                options.extend([
                    "-extraction_tech_file", self.get_qrc_tech(), # TODO: this assumes only 1 exists in no corners case
                    "-default_power_voltage", str(VoltageValue(self.get_setting("vlsi.inputs.supplies.VDD")).value_in_units("V"))
                ])
                ts_output.append("set_pg_library_mode {}".format(" ".join(options)))
                ts_output.append("write_pg_library -out_dir {}".format(self.tech_lib_dir))

                # Next do stdcell library
                options[options.index("techonly")] = "stdcells"
                if len(stdfillers) > 0:
                    stdfillers_names = list(map(lambda f: str(f), stdfillers[0].name))
                    options.extend(["-filler_cells", "{{ {} }} ".format(" ".join(stdfillers_names))])
                if len(decaps) > 0:
                    decaps_names = list(map(lambda d: str(d), decaps[0].name))
                    options.extend(["-decap_cells", "{{ {} }}".format(" ".join(decaps_names))])
                spice_models = self.technology.read_libs([hammer_tech.filters.spice_model_file_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
                spice_corners = self.technology.read_libs([hammer_tech.filters.spice_model_lib_corner_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
                if len(spice_models) == 0:
                    self.logger.error("Must specify Spice model files in tech plugin to generate stdcell PG libraries! Skipping.")
                    return True
                else:
                    options.extend(["-spice_models", "{", " ".join(spice_models), "}"])
                    if len(spice_corners) > 0:
                        options.extend(["-spice_corners", "{{", "} {".join(spice_corners), "}}"])
                if len(decaps) > 0 and len(tech_lib_sp) == 0:
                    self.logger.error("Must have Spice netlists in tech plugin for decap characterization in stdcell PG library! Skipping.")
                    return True
                else:
                    options.extend(["-spice_subckts", "{{ {} }}".format(" ".join(tech_lib_sp))])
                ts_output.append("set_pg_library_mode {}".format(" ".join(options)))
                ts_output.append("write_pg_library -out_dir {}".format(self.stdcell_lib_dir))

            else:
                for corner in corners:
                    # Start with tech-only library
                    options = tech_options.copy()
                    options.extend([
                        "-extraction_tech_file", self.get_mmmc_qrc(corner), #TODO: QRC should be tied to stackup
                        "-default_power_voltage", str(corner.voltage.value),
                        "-temperature", str(corner.temp.value)
                    ])
                    ts_output.append("set_pg_library_mode {}".format(" ".join(options)))
                    ts_output.append("write_pg_library -out_dir {}".format(os.path.join(self.tech_lib_dir, corner.name)))

                    # Next do stdcell library
                    options[options.index("techonly")] = "stdcells"
                    if len(stdfillers) > 0:
                        stdfillers_names = list(map(lambda f: str(f), stdfillers[0].name))
                        options.extend(["-filler_cells", "{{ {} }} ".format(" ".join(stdfillers_names))])
                    if len(decaps) > 0:
                        decaps_names = list(map(lambda d: str(d), decaps[0].name))
                        options.extend(["-decap_cells", "{{ {} }}".format(" ".join(decaps_names))])
                    spice_models = self.get_mmmc_spice_models(corner)
                    spice_corners = self.get_mmmc_spice_corners(corner)
                    if len(spice_models) == 0:
                        self.logger.error("Must specify Spice model files in tech plugin to generate stdcell PG libraries! Skipping.")
                        return True
                    else:
                        options.extend(["-spice_models", "{", " ".join(spice_models), "}"])
                        if len(spice_corners) > 0:
                            options.extend(["-spice_corners", "{{", "} {".join(spice_corners), "}}"])
                    if len(decaps) > 0 and len(tech_lib_sp) == 0:
                        self.logger.error("Must have Spice netlists in tech plugin for decap characterization in stdcell PG library! Skipping.")
                        return True
                    else:
                        options.extend(["-spice_subckts", "{{ {} }}".format(" ".join(tech_lib_sp))])
                    ts_output.append("set_pg_library_mode {}".format(" ".join(options)))
                    ts_output.append("write_pg_library -out_dir {}".format(os.path.join(self.stdcell_lib_dir, corner.name)))

            ts_output.append("exit")
            self.write_contents_to_path("\n".join(ts_output), self.tech_stdcell_pgv_tcl)
            self.gen_tech_stdcell_pgv = True
            self.ran_tech_stdcell_pgv = True
        else:
            self.logger.info("techonly and stdcell PG libraries already generated, skipping...")
            self.ran_tech_stdcell_pgv = True

        if self.get_setting("power.voltus.macro_pgv"):
            m_output = base_cmds.copy()
            # Characterize macro libraries once, unless list of extra libraries has been modified/changed
            tech_lef = tech_lib_lefs[0]
            extra_lib_lefs = self.technology.read_libs([hammer_tech.filters.lef_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.extra_lib_filter())
            extra_lib_mtimes = list(map(lambda l: os.path.getmtime(l), extra_lib_lefs))
            extra_lib_lefs_mtimes = dict(zip(extra_lib_lefs, extra_lib_mtimes))
            extra_lib_lefs_json = os.path.join(self.run_dir, "extra_lib_lefs.json")
            extra_pg_libs = self.technology.read_libs([hammer_tech.filters.power_grid_library_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.extra_lib_filter())
            # TODO: Use some filters w/ LEFUtils to extract cells from LEFs, e.g. MacroSize instead of using name field
            named_extra_libs = list(filter(lambda l: l.library.name is not None and l.library.power_grid_library not in extra_pg_libs, self.technology.get_extra_libraries()))  # type: List[hammer_tech.ExtraLibrary]

            if not os.path.isdir(self.macro_lib_dir):
                self.logger.info("Characterizing macros for the first time...")
                # First time: characterize all cells
                # macros = list(map(lambda l: l.library.name, named_extra_libs))
                macros = [l.library.name for l in named_extra_libs if l.library.name is not None]
                in_place_unique(macros)
                self.macro_pgv_cells = macros

                # Write dict of extra library LEFs
                with open(extra_lib_lefs_json, "w") as f:
                    f.write(json.dumps(extra_lib_lefs_mtimes, cls=HammerJSONEncoder, indent=4))
            else:
                # Figure out which cells to re-characterize
                prior_extra_lib_lefs = {}  # type: Dict[str, str]
                if os.path.exists(extra_lib_lefs_json):
                    with open(extra_lib_lefs_json, "r") as f:
                        prior_extra_lib_lefs = json.loads(f.read())
                # Write updated dict of extra library LEFs
                with open(extra_lib_lefs_json, "w") as f:
                    f.write(json.dumps(extra_lib_lefs_mtimes, cls=HammerJSONEncoder, indent=4))
                # Get LEFs which have been created/modified, match cell names if provided
                # TODO: these types don't line up, doubtful whether this code works as expected
                mod_lefs = dict(set(extra_lib_lefs_mtimes.items()) - set(prior_extra_lib_lefs.items())).keys() # type: ignore
                rechar_libs = list(filter(lambda l: l.library.lef_file in mod_lefs, named_extra_libs))
                macros = list(map(lambda l: l.library.name, rechar_libs)) # type: ignore
                in_place_unique(macros)
                self.macro_pgv_cells = macros

            if len(self.macro_pgv_cells) > 0:
                self.logger.info("Characterizing the following macros: {}".format(" ".join(self.macro_pgv_cells)))
                # Write list of cells to characterize
                cells_list = os.path.join(self.run_dir, "macro_cells.txt")
                with open(cells_list, "w") as f:
                    f.write("\n".join(self.macro_pgv_cells))

                macro_options = base_options.copy()
                macro_options.extend([
                        "-cell_type", "macros",
                        "-cells_file", cells_list
                ])

                # File checks
                gds_map_file = self.get_gds_map_file()
                if gds_map_file is None:
                    self.logger.error("Must have GDS layer map for macro PG library generation! Skipping.")
                    return True
                else:
                    assert isinstance(gds_map_file, str)
                    macro_options.extend(["-stream_layer_map", gds_map_file])

                extra_lib_sp = self.technology.read_libs([hammer_tech.filters.spice_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.extra_lib_filter())
                if len(extra_lib_sp) == 0:
                    self.logger.error("Must have Spice netlists for macro PG library generation! Skipping.")
                    return True
                else:
                    macro_options.extend(["-spice_subckts", "{{ {} }}".format(" ".join(extra_lib_sp))])

                extra_lib_gds = self.technology.read_libs([hammer_tech.filters.gds_filter], hammer_tech.HammerTechnologyUtils.to_plain_item, self.extra_lib_filter())
                if len(extra_lib_gds) == 0:
                    self.logger.error("Must have GDS data for macro PG library generation! Skipping.")
                    return True
                else:
                    macro_options.extend(["-stream_files", "{{ {} }}".format(" ".join(extra_lib_gds))])

                m_output.append("read_physical -lef {{ {TECH_LEF} {EXTRA_LEFS} }}".format(TECH_LEF=tech_lef, EXTRA_LEFS=" ".join(extra_lib_lefs)))

                if not corners:
                    options = macro_options.copy()
                    options.extend([
                        "-extraction_tech_file", self.get_qrc_tech(), # TODO: this assumes only 1 exists in no corners case
                        "-default_power_voltage", str(VoltageValue(self.get_setting("vlsi.inputs.supplies.VDD")).value_in_units("V"))
                    ])
                    spice_models = self.technology.read_libs([hammer_tech.filters.spice_model_file_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
                    spice_corners = self.technology.read_libs([hammer_tech.filters.spice_model_lib_corner_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
                    if len(spice_models) == 0:
                        self.logger.error("Must specify Spice model files in tech plugin to generate macro PG libraries")
                        return True
                    else:
                        options.extend(["-spice_models", "{", " ".join(spice_models), "}"])
                        if len(spice_corners) > 0:
                            options.extend(["-spice_corners", "{{", "} {".join(spice_corners), "}}"])
                    m_output.append("set_pg_library_mode {}".format(" ".join(options)))
                    m_output.append("write_pg_library -out_dir {}".format(os.path.join(self.macro_lib_dir, corner.name)))

                else:
                    for corner in corners:
                        options = macro_options.copy()
                        options.extend([
                            "-extraction_tech_file", self.get_mmmc_qrc(corner), #TODO: QRC should be tied to stackup
                            "-default_power_voltage", str(corner.voltage.value),
                            "-temperature", str(corner.temp.value),
                        ])
                        spice_models = self.get_mmmc_spice_models(corner)
                        spice_corners = self.get_mmmc_spice_corners(corner)
                        if len(spice_models) == 0:
                            self.logger.error("Must specify Spice model files in tech plugin to generate macro PG libraries")
                            return True
                        else:
                            options.extend(["-spice_models", "{", " ".join(spice_models), "}"])
                            if len(spice_corners) > 0:
                                options.extend(["-spice_corners", "{{", "} {".join(spice_corners), "}}"])
                        m_output.append("set_pg_library_mode {}".format(" ".join(options)))
                        m_output.append("write_pg_library -out_dir {}".format(os.path.join(self.macro_lib_dir, corner.name)))

                m_output.append("exit")
                self.write_contents_to_path("\n".join(m_output), self.macro_pgv_tcl)
                self.gen_macro_pgv = True
                self.ran_macro_pgv = True
            else:
                self.logger.info("macro PG libraries already generated and macros have not changed, skipping...")
                self.ran_macro_pgv = True
        else:
            self.logger.info("power.voltus.macro_pgv is False. Rail analysis will be incomplete over macro blocks.")

        return True

    def init_design(self) -> bool:
        verbose_append = self.verbose_append

        verbose_append("set_db design_process_node {}".format(self.get_setting("vlsi.core.node")))
        verbose_append("set_multi_cpu_usage -local_cpu {}".format(self.get_setting("vlsi.core.max_threads")))

        innovus_db = os.path.join(os.getcwd(), self.flow_database)
        if innovus_db is None or not os.path.isdir(innovus_db):
            raise ValueError("Innovus database %s not found" % (innovus_db))

        verbose_append("read_db {}".format(innovus_db))

        verbose_append("check_pg_shorts -out_file shorts.rpt")

        # TODO (daniel) deal with multiple power domains
        for power_net in self.get_all_power_nets():
            vdd_net = power_net.name
        for gnd_net in self.get_all_ground_nets():
            vss_net = gnd_net.name
        verbose_append("set_power_pads -net {VDD} -format defpin".format(VDD=vdd_net))
        verbose_append("set_power_pads -net {VSS} -format defpin".format(VSS=vss_net))

        # Check that SPEFs exist
        if len(self.spefs) == 0:
            self.logger.error("No spef files specified for power analysis")
            return False

        corners = self.get_mmmc_corners()
        if corners:
            setup_view_names = [] # type: List[str]
            hold_view_names = [] # type: List[str]
            extra_view_names = [] # type: List[str]
            rc_corners = [] # type: List[str]
            for corner in corners:
                # Setting up views for all defined corner types: setup, hold, extra
                if corner.type is MMMCCornerType.Setup:
                    corner_name = "{n}.{t}".format(n=corner.name, t="setup")
                    setup_view_names.append("{n}_view".format(n=corner_name))
                elif corner.type is MMMCCornerType.Hold:
                    corner_name = "{n}.{t}".format(n=corner.name, t="hold")
                    hold_view_names.append("{n}_view".format(n=corner_name))
                elif corner.type is MMMCCornerType.Extra:
                    corner_name = "{n}.{t}".format(n=corner.name, t="extra")
                    extra_view_names.append("{n}_view".format(n=corner_name))
                else:
                    raise ValueError("Unsupported MMMCCornerType")
                rc_corners.append("{n}_rc".format(n=corner_name))

            # Apply analysis views
            # TODO: should not need to analyze extra views as well. Defaulting to hold for now (min. runtime impact).
            verbose_append("set_analysis_view -setup {{ {setup_views} }} -hold {{ {hold_views} {extra_views} }}".format(
                setup_views=" ".join(setup_view_names),
                hold_views=" ".join(hold_view_names),
                extra_views=" ".join(extra_view_names)
            ))
            # Match spefs with corners. Ordering must match (ensured here by get_mmmc_corners())!
            for (spef, rc_corner) in zip(self.spefs, rc_corners):
                verbose_append("read_spef {spef} -rc_corner {corner}".format(spef=os.path.join(os.getcwd(), spef), corner=rc_corner))

        else:
            # TODO: remove hardcoded my_view string
            analysis_view_name = "my_view"
            verbose_append("set_analysis_view -setup {{ {setup_view} }} -hold {{ {hold_view} }}".format(
                setup_view=analysis_view_name,
                hold_view=analysis_view_name
            ))
            verbose_append("read_spef " + os.path.join(os.getcwd(), self.spefs[0]))

        return True

    def static_power(self) -> bool:
        verbose_append = self.verbose_append

        verbose_append("set_db power_method static")
        verbose_append("set_db power_write_static_currents true")
        verbose_append("set_db power_write_db true")

        # Report based on MMMC mode
        corners = self.get_mmmc_corners()
        if not corners:
            if self.extra_corners_only:
                self.logger.warning("power.inputs.extra_corners_only not valid in non-MMMC mode! Reporting static power for default analysis view only.")
            verbose_append("report_power -out_dir staticPowerReports")
        else:
            if self.extra_corners_only:
                extra_corners = list(filter(lambda c: c.type is MMMCCornerType.Extra, corners))
                if len(extra_corners) == 0:
                    self.logger.warning("power.inputs.extra_corners_only is true but no extra MMMC corners specified! Ignoring for static power.")
                else:
                    corners = extra_corners
            for corner in corners:
                # Setting up views for all defined corner types: setup, hold, extra
                if corner.type is MMMCCornerType.Setup:
                    view_name = "{c}.setup_view".format(c=corner.name)
                elif corner.type is MMMCCornerType.Hold:
                    view_name = "{c}.hold_view".format(c=corner.name)
                elif corner.type is MMMCCornerType.Extra:
                    view_name = "{c}.extra_view".format(c=corner.name)
                else:
                    raise ValueError("Unsupported MMMCCornerType")
                verbose_append("report_power -view {VIEW} -out_dir staticPowerReports.{VIEW}".format(VIEW=view_name))

        return True

    def active_power(self) -> bool:
        verbose_append = self.verbose_append

        # Active Vectorless Power Analysis
        verbose_append("set_db power_method dynamic_vectorless")
        # TODO (daniel) add the resolution as an option?
        verbose_append("set_dynamic_power_simulation -resolution 500ps")

        # Check MMMC mode
        corners = self.get_mmmc_corners()
        if not corners:
            if self.extra_corners_only:
                self.logger.warning("power.inputs.extra_corners_only not valid in non-MMMC mode! Reporting active power for default analysis view only.")
            verbose_append("report_power -out_dir activePowerReports")
        else:
            if self.extra_corners_only:
                extra_corners = list(filter(lambda c: c.type is MMMCCornerType.Extra, corners))
                if len(extra_corners) == 0:
                    self.logger.warning("power.inputs.extra_corners_only is true but no extra MMMC corners specified! Ignoring for active power.")
                else:
                    corners = extra_corners
            for corner in corners:
                # Setting up views for all defined corner types: setup, hold, extra
                if corner.type is MMMCCornerType.Setup:
                    view_name = "{c}.setup_view".format(c=corner.name)
                elif corner.type is MMMCCornerType.Hold:
                    view_name = "{c}.hold_view".format(c=corner.name)
                elif corner.type is MMMCCornerType.Extra:
                    view_name = "{c}.extra_view".format(c=corner.name)
                else:
                    raise ValueError("Unsupported MMMCCornerType")
                verbose_append("report_power -view {VIEW} -out_dir activePowerReports.{VIEW}".format(VIEW=view_name))

        # TODO (daniel) deal with different tb/dut hierarchies
        tb_name = self.get_setting("power.inputs.tb_name")
        tb_dut = self.get_setting("power.inputs.tb_dut")
        tb_scope = "{}/{}".format(tb_name, tb_dut)

        # TODO: These times should be either auto calculated/read from the inputs or moved into the same structure as a tuple
        start_times = self.get_setting("power.inputs.start_times")
        end_times = self.get_setting("power.inputs.end_times")

        # Active Vectorbased Power Analysis
        verbose_append("set_db power_method dynamic_vectorbased")
        waveform_format_map = {".vcd": "vcd",
                               ".vpd": "vcd",
                               ".fsdb": "fsdb",
                               ".shm": "shm",
                               ".trn": "shm"}
        for waveform_path, waveform_stime, waveform_etime in zip(self.waveforms, start_times, end_times):
            stime_ns = TimeValue(waveform_stime).value_in_units("ns")
            etime_ns = TimeValue(waveform_etime).value_in_units("ns")
            # Set format intelligently based on file extension. Strip .gz if present.
            waveform_ext = os.path.splitext(waveform_path.rstrip(".gz"))[1].lower()
            if waveform_format_map.get(waveform_ext) is None:
                self.logger.error("Only VCD/VPD, FSDB, and SHM waveform formats supported.")
            verbose_append("read_activity_file -reset -format {FORMAT} {WAVEFORM_PATH} -start {stime}ns -end {etime}ns -scope {TESTBENCH}".format(FORMAT=waveform_format_map.get(waveform_ext), WAVEFORM_PATH=os.path.join(os.getcwd(), waveform_path), TESTBENCH=tb_scope, stime=stime_ns, etime=etime_ns))
            waveform_file = os.path.basename(waveform_path)
            # Report based on MMMC mode
            if not corners:
                verbose_append("report_power -out_dir activePower.{WAVEFORM_FILE}".format(WAVEFORM_FILE=waveform_file))
            else:
                for corner in corners:
                    # Setting up views for all defined corner types: setup, hold, extra
                    if corner.type is MMMCCornerType.Setup:
                        view_name = "{c}.setup_view".format(c=corner.name)
                    elif corner.type is MMMCCornerType.Hold:
                        view_name = "{c}.hold_view".format(c=corner.name)
                    elif corner.type is MMMCCornerType.Extra:
                        view_name = "{c}.extra_view".format(c=corner.name)
                    else:
                        raise ValueError("Unsupported MMMCCornerType")
                    verbose_append("report_power -view {VIEW} -out_dir activePowerReports.{WAVEFORM_FILE}.{VIEW}".format(VIEW=view_name, WAVEFORM_FILE=waveform_file))

            verbose_append("report_vector_profile -detailed_report true -out_file activePowerProfile.{WAVEFORM_FILE}".format(WAVEFORM_FILE=waveform_file))

        verbose_append("set_db power_method dynamic")
        for saif_path in self.saifs:
            verbose_append("set_dynamic_power_simulation -reset")
            verbose_append("read_activity_file -reset -format SAIF {SAIF_PATH} -scope {TESTBENCH}".format(SAIF_PATH=os.path.join(os.getcwd(), saif_path), TESTBENCH=tb_scope))
            saif_file=".".join(saif_path.split('/')[-2:])
            # Report based on MMMC mode
            if not corners:
                verbose_append("report_power -out_dir activePower.{SAIF_FILE}".format(SAIF_FILE=saif_file))
            else:
                for corner in corners:
                    # Setting up views for all defined corner types: setup, hold, extra
                    if corner.type is MMMCCornerType.Setup:
                        view_name = "{c}.setup_view".format(c=corner.name)
                    elif corner.type is MMMCCornerType.Hold:
                        view_name = "{c}.hold_view".format(c=corner.name)
                    elif corner.type is MMMCCornerType.Extra:
                        view_name = "{c}.extra_view".format(c=corner.name)
                    else:
                        raise ValueError("Unsupported MMMCCornerType")
                    verbose_append("report_power -view {VIEW} -out_dir activePowerReports.{SAIF_FILE}.{VIEW}".format(VIEW=view_name, SAIF_FILE=saif_file))

        return True

    def rail_analysis(self, method: str, power_dir: str, output_dir: Optional[str] = None) -> bool:
        """
        Generic method for rail analysis
        params:
        - method: "static" or "dynamic"
        - power_dir: relative path to static or active power current files
        - output_dir: relative path to rail analysis output dir
        """
        verbose_append = self.verbose_append

        if not output_dir:
            output_dir = method + "RailReports"

        # Decide accuracy based on existence of PGV libraries, unless overridden
        accuracy = self.get_setting("power.voltus.rail_accuracy")
        if not accuracy:
            accuracy = "hd" if self.ran_tech_stdcell_pgv else "xd" # hd still works w/o macro PG views

        base_options = [
            "-method", method,
            "-accuracy", accuracy,
            "-process_techgen_em_rules", "true",
            "-em_peak_analysis", "true",
            "-enable_rlrp_analysis", "true",
            "-gif_resolution", "high",
            "-verbosity", "true"
        ]
        if method == "static":
            base_options.extend(["-enable_sensitivity_analysis", "true"])

        # TODO: Need combinations of all power nets + voltage domains
        pg_nets = self.get_all_power_nets() + self.get_all_ground_nets()
        # Report based on MMMC corners
        corners = self.get_mmmc_corners()
        if not corners:
            if self.extra_corners_only:
                self.logger.warning("power.inputs.extra_corners_only not valid in non-MMMC mode! Reporting rail analysis for default analysis view only.")
            options = base_options.copy()
            pg_libs = self.technology.read_libs([hammer_tech.filters.power_grid_library_filter], hammer_tech.HammerTechnologyUtils.to_plain_item)
            if self.ran_tech_stdcell_pgv:
                pg_libs.append(os.path.join(self.tech_lib_dir, "techonly.cl"))
                pg_libs.append(os.path.join(self.stdcell_lib_dir, "stdcells.cl"))
            if self.ran_macro_pgv:
                pg_libs.extend(list(map(lambda l: os.path.join(self.macro_lib_dir, "macros_{}.cl".format(l)), self.macro_pgv_cells)))
            if len(pg_libs) == 0:
                self.logger.warning("No PG libraries are available! Rail analysis is skipped.")
                return True
            options.extend(["-power_grid_libraries", "{{ {} }}".format(" ".join(pg_libs))])
            verbose_append("set_rail_analysis_config {}".format(" ".join(options)))
            # TODO: get nets and .ptiavg files using TCL from the .ptifiles file in the power reports directory
            power_data = list(map(lambda n: "{POWER_DIR}/{METHOD}_{NET}.ptiavg".format(
                POWER_DIR=power_dir,
                METHOD=method,
                NET=n.name), pg_nets))
            verbose_append("set_power_data -format current {{ {} }}".format(" ".join(power_data)))
            verbose_append("report_rail -output_dir {} -type domain ALL".format(output_dir))
            # TODO: Find highest run number, increment by 1 to enable reporting IRdrop regions
        else:
            if self.extra_corners_only:
                extra_corners = list(filter(lambda c: c.type is MMMCCornerType.Extra, corners))
                if len(extra_corners) == 0:
                    self.logger.warning("power.inputs.extra_corners_only is true but no extra MMMC corners specified! Ignoring for rail analysis.")
                else:
                    corners = extra_corners
            for corner in corners:
                options = base_options.copy()
                if corner.type is MMMCCornerType.Setup:
                    view_name = corner.name + ".setup_view"
                elif corner.type is MMMCCornerType.Hold:
                    view_name = corner.name + ".hold_view"
                elif corner.type is MMMCCornerType.Extra:
                    view_name = corner.name + ".extra_view"
                else:
                    raise ValueError("Unsupported MMMCCornerType")
                pg_libs = self.get_mmmc_pgv(corner)
                if self.ran_tech_stdcell_pgv:
                    pg_libs.append(os.path.join(self.tech_lib_dir, corner.name, "techonly.cl"))
                    pg_libs.append(os.path.join(self.stdcell_lib_dir, corner.name, "stdcells.cl"))
                if self.ran_macro_pgv:
                    pg_libs.extend(list(map(lambda l: os.path.join(self.macro_lib_dir, corner.name, "macros_{}.cl".format(l)), self.macro_pgv_cells)))
                if len(pg_libs) == 0:
                    self.logger.warning("No PG libraries are available! Rail analysis is skipped.")
                    return True

                options.extend([
                    "-power_grid_libraries", "{{ {} }}".format(" ".join(pg_libs)),
                    "-analysis_view", view_name,
                    "-temperature", str(corner.temp.value)
                ])
                verbose_append("set_rail_analysis_config {}".format(" ".join(options)))
                verbose_append("set_power_data -reset")
                # TODO: get nets and .ptiavg files using TCL from the .ptifiles file in the power reports directory
                power_data = list(map(lambda n: "{POWER_DIR}.{VIEW}/{METHOD}_{NET}.ptiavg".format(
                    POWER_DIR=power_dir,
                    VIEW=view_name,
                    METHOD=method,
                    NET=n.name), pg_nets))
                verbose_append("set_power_data -format current {{ {} }}".format(" ".join(power_data)))
                verbose_append("report_rail -output_dir {} -type domain ALL".format(output_dir))
                # TODO: Find highest run number, increment by 1 to enable reporting IRdrop regions

        return True

    def static_rail(self) -> bool:
        return self.rail_analysis("static", "staticPowerReports")

    def active_rail(self) -> bool:
        # Vectorless database
        passed = self.rail_analysis("dynamic", "activePowerReports", "activeRailReports")

        # Vectorbased databases
        for waveform_path in self.waveforms:
            passed = self.rail_analysis("dynamic", "activePower." + os.path.basename(waveform_path), "activeRailReports." + os.path.basename(waveform_path))
        for saif_path in self.saifs:
            saif_file=".".join(saif_path.split('/')[-2:])
            passed = self.rail_analysis("dynamic", "activePower." + saif_file, "activeRailReports." + saif_file)
        return passed

    def run_voltus(self) -> bool:
        verbose_append = self.verbose_append

        """Close out the power script and run Voltus"""
        # Quit Voltus
        verbose_append("exit")

        # Create power analysis script
        power_tcl_filename = os.path.join(self.run_dir, "power.tcl")
        self.write_contents_to_path("\n".join(self.output), power_tcl_filename)

        # Build args
        base_args = [
            self.get_setting("power.voltus.voltus_bin"),
            "-no_gui",
            "-common_ui",
            "-init"
        ]

        HammerVLSILogging.enable_colour = False
        HammerVLSILogging.enable_tag = False

        # Run PG lib gen, if needed
        if self.gen_tech_stdcell_pgv:
            args = base_args.copy()
            args.append(self.tech_stdcell_pgv_tcl)
            self.run_executable(args, cwd=self.run_dir)
        if self.gen_macro_pgv:
            args = base_args.copy()
            args.append(self.macro_pgv_tcl)
            self.run_executable(args, cwd=self.run_dir)

        args = base_args.copy()
        args.append(power_tcl_filename)
        self.run_executable(args, cwd=self.run_dir)

        HammerVLSILogging.enable_colour = True
        HammerVLSILogging.enable_tag = True

        return True



tool = Voltus



#  nop.py
#  No-op place and route tool.
#
#  See LICENSE for licence details.

from hammer.vlsi import HammerPlaceAndRouteTool, DummyHammerTool
from typing import List, Optional
from decimal import Decimal


class NopPlaceAndRoute(HammerPlaceAndRouteTool, DummyHammerTool):
    def fill_outputs(self) -> bool:
        self.output_ilms = []
        self.output_gds = "/dev/null"
        self.output_netlist = "/dev/null"
        self.output_sim_netlist = "/dev/null"
        self.hcells_list = []
        return True

    def specify_power_straps(self, layer_name: str, bottom_via_layer_name: str, blockage_spacing: Decimal, pitch: Decimal, width: Decimal, spacing: Decimal, offset: Decimal, bbox: Optional[List[Decimal]], nets: List[str], add_pins: bool) -> List[str]:
        return []

    def specify_std_cell_power_straps(self, blockage_spacing: Decimal, bbox: Optional[List[Decimal]], nets: List[str]) -> List[str]:
        return []


tool = NopPlaceAndRoute



#  Vivado synthesis plugin for Hammer.
#
#  See LICENSE for licence details.

from typing import List, Optional

import os

from hammer.vlsi import HammerSynthesisTool, HammerToolStep

from hammer.common.vivado.vivado_core import VivadoCommon


class VivadoSynth(HammerSynthesisTool, VivadoCommon):
    @property
    def steps(self) -> List[HammerToolStep]:
        return self.make_steps_from_methods([
            self.setup_workspace,
            self.generate_board_defs,
            self.generate_paths_and_src_defs,
            self.generate_project_defs,
            self.generate_prologue,
            self.generate_ip_defs,
            self.generate_messaging_params,
            self.generate_synth_cmds,
            self.run_synthesis,
        ])

    def generate_prologue(self) -> bool:
        self.append_file('prologue.tcl', None)
        return True

    def generate_ip_defs(self) -> bool:
        ip_file = self.get_setting('synthesis.vivado.ip_def_tcl')  # type: Optional[str]
        if ip_file:
            self.append_file(ip_file, None)
        return True

    def generate_messaging_params(self) -> bool:
        self.append_file('msg.tcl', None)
        return True

    def generate_synth_cmds(self) -> bool:
        ip_file = self.get_setting('synthesis.vivado.ip_def_tcl')
        if ip_file:
            self.append(
                'read_ip [glob -directory $ipdir [file join * {*.xci}]]')
        self.append_file('syn.tcl', None)
        return True

    def run_synthesis(self) -> bool:
        # Create synthesis script.
        syn_tcl_filename = os.path.join(self.run_dir, "syn.tcl")

        with open(syn_tcl_filename, "w") as f:
            f.write("\n".join(self.output))

        # create executable
        file_params = {
            'env_setup_script':
            self.get_setting('synthesis.vivado.setup_script'),
            'work_dir': self.run_dir,
            'vivado_cmd': self.get_setting('synthesis.vivado.binary'),
        }
        run_script = self.generate_run_script('run-synthesis', file_params)

        # run executable
        self.run_executable([run_script])
        return True

    def fill_outputs(self) -> bool:
        dcp_path = os.path.join(self.run_dir, 'obj', 'post_opt.dcp')
        dcp_dir = os.path.dirname(dcp_path)
        self.output_files = [
            dcp_path,
            os.path.join(dcp_dir, self.top_module + '_post_synth.v')
        ]
        return True


tool = VivadoSynth
